import os
import glob

for gd_mod in glob.glob("src/*/"):
    mod_files = []
    for paths in [gd_mod + 'utilities', gd_mod + 'example']:
        if os.path.isdir(paths):
            for root, dirs, files in os.walk(paths):
                for file in files:
                    if file.endswith(".cpp"):
                        mod_files.append(str(os.path.join(root, file)).split(paths)[1][1:])
    if len(mod_files) > 0:
        mod = str(gd_mod).split('/')[1]
        print(' * \section ' + mod + '_example_section ' + mod)
        for file in mod_files:
            print(' * @example ' + file)



import numpy               as np
import tensorflow          as tf
from ..cubical_complex  import CubicalComplex

######################
# Cubical filtration #
######################

# The parameters of the model are the pixel values.

def _Cubical(Xflat, Xdim, dimensions, homology_coeff_field):
    # Parameters: Xflat (flattened image),
    #             Xdim (shape of non-flattened image)
    #             dimensions (homology dimensions)

    # Compute the persistence pairs with Gudhi
    # We reverse the dimensions because CubicalComplex uses Fortran ordering
    cc = CubicalComplex(dimensions=Xdim[::-1], top_dimensional_cells=Xflat)
    cc.compute_persistence(homology_coeff_field=homology_coeff_field)

    # Retrieve and output image indices/pixels corresponding to positive and negative simplices
    cof_pp = cc.cofaces_of_persistence_pairs()
    
    L_cofs = []
    for dim in dimensions:

        try:
            cof = cof_pp[0][dim]
        except IndexError:
            cof = np.array([])

        L_cofs.append(np.array(cof, dtype=np.int32))

    return L_cofs

class CubicalLayer(tf.keras.layers.Layer):
    """
    TensorFlow layer for computing the persistent homology of a cubical complex
    """
    def __init__(self, homology_dimensions, min_persistence=None, homology_coeff_field=11, **kwargs):
        """
        Constructor for the CubicalLayer class

        Parameters:
            homology_dimensions (List[int]): list of homology dimensions
            min_persistence (List[float]): minimum distance-to-diagonal of the points in the output persistence diagrams (default None, in which case 0. is used for all dimensions)
            homology_coeff_field (int): homology field coefficient. Must be a prime number. Default value is 11. Max is 46337.
        """
        super().__init__(**kwargs)
        self.dimensions = homology_dimensions
        self.min_persistence = min_persistence if min_persistence is not None else [0.] * len(self.dimensions)
        self.hcf = homology_coeff_field
        assert len(self.min_persistence) == len(self.dimensions)

    def call(self, X):
        """
        Compute persistence diagram associated to a cubical complex filtered by some pixel values 

        Parameters:
            X (TensorFlow variable): pixel values of the cubical complex

        Returns:
            List[Tuple[tf.Tensor,tf.Tensor]]: List of cubical persistence diagrams. The length of this list is the same than that of dimensions, i.e., there is one persistence diagram per homology dimension provided in the input list dimensions. Moreover, the finite and essential parts of the persistence diagrams are provided separately: each element of this list is a tuple of size two that contains the finite and essential parts of the corresponding persistence diagram, of shapes [num_finite_points, 2] and [num_essential_points, 1] respectively. Note that the essential part is always empty in cubical persistence diagrams, except in homology dimension zero, where the essential part always contains a single point, with abscissa equal to the smallest value in the complex, and infinite ordinate
        """
        # Compute pixels associated to positive and negative simplices 
        # Don't compute gradient for this operation
        Xflat = tf.reshape(X, [-1])
        Xdim, Xflat_numpy = X.shape, Xflat.numpy()
        indices_list = _Cubical(Xflat_numpy, Xdim, self.dimensions, self.hcf)
        index_essential = np.argmin(Xflat_numpy) # index of minimum pixel value for essential persistence diagram
        # Get persistence diagram by simply picking the corresponding entries in the image
        self.dgms = []
        for idx_dim, dimension in enumerate(self.dimensions):
            finite_dgm = tf.reshape(tf.gather(Xflat, indices_list[idx_dim]), [-1,2])
            essential_dgm = tf.reshape(tf.gather(Xflat, index_essential), [-1,1]) if dimension == 0 else tf.zeros([0, 1])
            min_pers = self.min_persistence[idx_dim]
            if min_pers >= 0:
                persistent_indices = tf.where(tf.math.abs(finite_dgm[:,1]-finite_dgm[:,0]) > min_pers)
                self.dgms.append((tf.reshape(tf.gather(finite_dgm, indices=persistent_indices), [-1,2]), essential_dgm))
            else:
                self.dgms.append((finite_dgm, essential_dgm))
        return self.dgms



#!/usr/bin/env python

import argparse
import errno
import os
import gudhi

""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ -
    which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full
    license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2016 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

parser = argparse.ArgumentParser(
    description="EuclideanWitnessComplex creation from " "points read in a OFF file.",
    epilog="Example: "
    "example/euclidean_witness_complex_diagram_persistence_from_off_file_example.py "
    "-f ../data/points/tore3D_300.off -a 1.0 -n 20 -d 2"
    "- Constructs a weak witness complex with the "
    "points from the given OFF file.",
)
parser.add_argument("-f", "--file", type=str, required=True)
parser.add_argument("-a", "--max_alpha_square", type=float, required=True)
parser.add_argument("-n", "--number_of_landmarks", type=int, required=True)
parser.add_argument("-d", "--limit_dimension", type=int, required=True)
parser.add_argument("-b", "--band", type=float, default=0.0)
parser.add_argument(
    "--no-diagram",
    default=False,
    action="store_true",
    help="Flag for not to display the diagrams",
)

args = parser.parse_args()

with open(args.file) as f:
    first_line = f.readline()
    if (first_line == "OFF\n") or (first_line == "nOFF\n"):
        print("#####################################################################")
        print("EuclideanWitnessComplex creation from points read in a OFF file")

        witnesses = gudhi.read_points_from_off_file(off_file=args.file)
        landmarks = gudhi.pick_n_random_points(
            points=witnesses, nb_points=args.number_of_landmarks
        )

        message = (
            "EuclideanWitnessComplex with max_edge_length="
            + repr(args.max_alpha_square)
            + " - Number of landmarks="
            + repr(args.number_of_landmarks)
        )
        print(message)

        witness_complex = gudhi.EuclideanWitnessComplex(
            witnesses=witnesses, landmarks=landmarks
        )
        simplex_tree = witness_complex.create_simplex_tree(
            max_alpha_square=args.max_alpha_square, limit_dimension=args.limit_dimension
        )

        message = "Number of simplices=" + repr(simplex_tree.num_simplices())
        print(message)

        diag = simplex_tree.persistence()

        print("betti_numbers()=")
        print(simplex_tree.betti_numbers())

        if args.no_diagram == False:
            import matplotlib.pyplot as plot
            gudhi.plot_persistence_diagram(diag, band=args.band)
            plot.show()
    else:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                args.file)

    f.close()



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Marc Glisse
#
# Copyright (C) 2023 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from .._edge_collapse import _collapse_edges


def reduce_graph(input_edges, nb_iterations=1):
    """Simplify a clique filtration, preserving its persistent homology.
    The clique filtration is represented as a sparse weighted graph, giving for each edge its 2 vertices
    and its filtration value, and the output is in the same format. An edge may be represented arbitrarily as (i, j)
    or (j, i). Listing the same edge twice, or a self-loop, is undefined.
    The cliques of the graph composed of the edges with filtration value less than some arbitrary value implicitly
    define a flag complex, so the weighted graph describes a flag filtration.
    This function outputs another flag filtration which is guaranteed to have the same persistent homology as the input.
    The algorithm works by collapsing edges, as described in :cite:`edgecollapsearxiv`.
    Note that this simplification is independent of the filtration values of the vertices.
    The output has the same shape as the input, which is presumed to be (N, N) where all vertices have index
    less than N, since the simplification does not affect vertices.

    :param input_edges: Input weighted graph.
    :type input_edges: scipy.sparse.coo_matrix
    :param nb_iterations: The number of times we apply the algorithm. Default is 1.
    :type nb_iterations: int
    :rtype: scipy.sparse.coo_matrix
    """
    from scipy.sparse import coo_matrix

    r = _collapse_edges(input_edges.row, input_edges.col, input_edges.data, nb_iterations)
    return coo_matrix((r[1], (r[0][0], r[0][1])), input_edges.shape)



""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2021 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

import gudhi as gd
import numpy as np
import matplotlib as plt
import pytest
import warnings


def test_array_handler():
    diags = np.array([[1, 2], [3, 4], [5, 6]], float)
    arr_diags, nx2_array = gd.persistence_graphical_tools._array_handler(diags)
    assert nx2_array
    for idx in range(len(diags)):
        assert arr_diags[idx][0] == 0
        np.testing.assert_array_equal(arr_diags[idx][1], diags[idx])

    diags = [(1.0, 2.0), (3.0, 4.0), (5.0, 6.0)]
    arr_diags, nx2_array = gd.persistence_graphical_tools._array_handler(diags)
    assert nx2_array
    for idx in range(len(diags)):
        assert arr_diags[idx][0] == 0
        assert arr_diags[idx][1] == diags[idx]

    diags = [(0, (1.0, 2.0)), (0, (3.0, 4.0)), (0, (5.0, 6.0))]
    arr_diags, nx2_array = gd.persistence_graphical_tools._array_handler(diags)
    assert not nx2_array
    assert arr_diags == diags


def test_min_birth_max_death():
    diags = [
        (0, (0.0, float("inf"))),
        (0, (0.0983494, float("inf"))),
        (0, (0.0, 0.122545)),
        (0, (0.0, 0.12047)),
        (0, (0.0, 0.118398)),
        (0, (0.118398, 1.0)),
        (0, (0.0, 0.117908)),
        (0, (0.0, 0.112307)),
        (0, (0.0, 0.107535)),
        (0, (0.0, 0.106382)),
    ]
    assert gd.persistence_graphical_tools._min_birth_max_death(diags) == (0.0, 1.0)
    assert gd.persistence_graphical_tools._min_birth_max_death(diags, band=4.0) == (0.0, 5.0)


def test_limit_min_birth_max_death():
    diags = [
        (0, (2.0, float("inf"))),
        (0, (2.0, float("inf"))),
    ]
    assert gd.persistence_graphical_tools._min_birth_max_death(diags) == (2.0, 3.0)
    assert gd.persistence_graphical_tools._min_birth_max_death(diags, band=4.0) == (2.0, 6.0)


def test_limit_to_max_intervals():
    diags = [
        (0, (0.0, float("inf"))),
        (0, (0.0983494, float("inf"))),
        (0, (0.0, 0.122545)),
        (0, (0.0, 0.12047)),
        (0, (0.0, 0.118398)),
        (0, (0.118398, 1.0)),
        (0, (0.0, 0.117908)),
        (0, (0.0, 0.112307)),
        (0, (0.0, 0.107535)),
        (0, (0.0, 0.106382)),
    ]
    # check no warnings if max_intervals equals to the diagrams number
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        truncated_diags = gd.persistence_graphical_tools._limit_to_max_intervals(
            diags, 10, key=lambda life_time: life_time[1][1] - life_time[1][0]
        )
        # check diagrams are not sorted
        assert truncated_diags == diags

    # check warning if max_intervals lower than the diagrams number
    with pytest.warns(UserWarning) as record:
        truncated_diags = gd.persistence_graphical_tools._limit_to_max_intervals(
            diags, 5, key=lambda life_time: life_time[1][1] - life_time[1][0]
        )
        # check diagrams are truncated and sorted by life time
        assert truncated_diags == [
            (0, (0.0, float("inf"))),
            (0, (0.0983494, float("inf"))),
            (0, (0.118398, 1.0)),
            (0, (0.0, 0.122545)),
            (0, (0.0, 0.12047)),
        ]
    assert len(record) == 1


def _limit_plot_persistence(function):
    pplot = function(persistence=[])
    assert isinstance(pplot, plt.axes.SubplotBase)
    pplot = function(persistence=[], legend=True)
    assert isinstance(pplot, plt.axes.SubplotBase)
    pplot = function(persistence=[(0, float("inf"))])
    assert isinstance(pplot, plt.axes.SubplotBase)
    pplot = function(persistence=[(0, float("inf"))], legend=True)
    assert isinstance(pplot, plt.axes.SubplotBase)


def test_limit_plot_persistence():
    for function in [gd.plot_persistence_barcode, gd.plot_persistence_diagram, gd.plot_persistence_density]:
        _limit_plot_persistence(function)


def _non_existing_persistence_file(function):
    with pytest.raises(FileNotFoundError):
        function(persistence_file="pouetpouettralala.toubiloubabdou")


def test_non_existing_persistence_file():
    for function in [gd.plot_persistence_barcode, gd.plot_persistence_diagram, gd.plot_persistence_density]:
        _non_existing_persistence_file(function)



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Marc Glisse
#
# Copyright (C) 2020 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

import numpy
from ..point_cloud.knn import KNearestNeighbors
from ..point_cloud.dtm import DTMDensity
from ._tomato import *

# The fit/predict interface is not so well suited...


class Tomato:
    """
    This clustering algorithm needs a neighborhood graph on the points, and an estimation of the density at each point.
    A few possible graph constructions and density estimators are provided for convenience, but it is perfectly natural
    to provide your own.

    :Requires: `SciPy <installation.html#scipy>`_, `Scikit-learn <installation.html#scikit-learn>`_ or others
        (see :class:`~gudhi.point_cloud.knn.KNearestNeighbors`) in function of the options.

    Attributes
    ----------
    n_clusters_: int
        The number of clusters. Writing to it automatically adjusts `labels_`.
    merge_threshold_: float
        minimum prominence of a cluster so it doesn't get merged. Writing to it automatically adjusts `labels_`.
    n_leaves_: int
        number of leaves (unstable clusters) in the hierarchical tree
    leaf_labels_: ndarray of shape (n_samples,)
        cluster labels for each point, at the very bottom of the hierarchy
    labels_: ndarray of shape (n_samples,)
        cluster labels for each point, after merging
    diagram_: ndarray of shape (`n_leaves_`, 2)
        persistence diagram (only the finite points)
    max_weight_per_cc_: ndarray of shape (n_connected_components,)
        maximum of the density function on each connected component. This corresponds to the abscissa of infinite
        points in the diagram
    children_: ndarray of shape (`n_leaves_`-n_connected_components, 2)
        The children of each non-leaf node. Values less than `n_leaves_` correspond to leaves of the tree.
        A node i greater than or equal to `n_leaves_` is a non-leaf node and has children children_[i - `n_leaves_`].
        Alternatively at the i-th iteration, children[i][0] and children[i][1] are merged to form node `n_leaves_` + i
    weights_: ndarray of shape (n_samples,)
        weights of the points, as computed by the density estimator or provided by the user
    params_: dict
        Parameters like metric, etc
    """

    def __init__(
        self,
        graph_type="knn",
        density_type="logDTM",
        n_clusters=None,
        merge_threshold=None,
        #       eliminate_threshold=None,
        #           eliminate_threshold (float): minimum max weight of a cluster so it doesn't get eliminated
        **params
    ):
        """
        Args:
            graph_type (str): 'manual', 'knn' or 'radius'. Default is 'knn'.
            density_type (str): 'manual', 'DTM', 'logDTM', 'KDE' or 'logKDE'. When you have many points,
                'KDE' and 'logKDE' tend to be slower. Default is 'logDTM'.
                The values computed for 'DTM' or 'KDE' are not normalized (this does not affect the clustering).
            metric (str|Callable): metric used when calculating the distance between instances in a feature array.
                Defaults to Minkowski of parameter p.
            kde_params (dict): if density_type is 'KDE' or 'logKDE', additional parameters passed directly to
                sklearn.neighbors.KernelDensity.
            k (int): number of neighbors for a knn graph (including the vertex itself). Defaults to 10.
            k_DTM (int): number of neighbors for the DTM density estimation (including the vertex itself).
                Defaults to k.
            r (float): size of a neighborhood if graph_type is 'radius'. Also used as default bandwidth in kde_params.
            eps (float): (1+eps) approximation factor when computing distances (ignored in many cases).
            n_clusters (int): number of clusters requested. Defaults to None, i.e. no merging occurs and we get
                the maximal number of clusters.
            merge_threshold (float): minimum prominence of a cluster so it doesn't get merged.
            symmetrize_graph (bool): whether we should add edges to make the neighborhood graph symmetric.
                This can be useful with k-NN for small k. Defaults to false.
            p (float): norm L^p on input points. Defaults to 2.
            q (float): order used to compute the distance to measure. Defaults to dim.
            dim (float): final exponent in DTM density estimation, representing the dimension. Defaults to the
                dimension, or 2 when the dimension cannot be read from the input (metric is "precomputed").
                Beware that when the dimension is large, this can easily cause overflows, see
                :class:`~gudhi.point_cloud.dtm.DTMDensity` for details
            n_jobs (int): Number of jobs to schedule for parallel processing on the CPU.
                If -1 is given all processors are used. Default: 1.
            params: extra parameters are passed to :class:`~gudhi.point_cloud.knn.KNearestNeighbors` and
                :class:`~gudhi.point_cloud.dtm.DTMDensity`.
        """
        # Should metric='precomputed' mean input_type='distance_matrix'?
        # Should we be able to pass metric='minkowski' (what None does currently)?
        self.graph_type_ = graph_type
        self.density_type_ = density_type
        self.params_ = params
        self.__n_clusters = n_clusters
        self.__merge_threshold = merge_threshold
        # self.eliminate_threshold_ = eliminate_threshold
        if n_clusters and merge_threshold:
            raise ValueError("Cannot specify both a merge threshold and a number of clusters")

    def fit(self, X, y=None, weights=None):
        """
        Args:
            X ((n,d)-array of float|(n,n)-array of float|Sequence[Iterable[int]]): coordinates of the points,
                or distance matrix (full, not just a triangle) if metric is "precomputed", or list of neighbors
                for each point (points are represented by their index, starting from 0) if graph_type is "manual".
                The number of points is currently limited to about 2 billion.
            weights (ndarray of shape (n_samples)): if density_type is 'manual', a density estimate at each point
            y: Not used, present here for API consistency with scikit-learn by convention.
        """
        # TODO: First detect if this is a new call with the same data (only threshold changed?)
        # TODO: less code duplication (subroutines?), less spaghetti, but don't compute neighbors twice if not needed. Clear error message for missing or contradictory parameters.
        if weights is not None:
            density_type = "manual"
        else:
            density_type = self.density_type_
            if density_type == "manual":
                raise ValueError("If density_type is 'manual', you must provide weights to fit()")

        if self.graph_type_ == "manual":
            self.neighbors_ = X
            # FIXME: uniformize "message 'option'" vs 'message "option"'
            assert density_type == "manual", 'If graph_type is "manual", density_type must be as well'
        else:
            metric = self.params_.get("metric", "minkowski")
            if metric != "precomputed":
                self.points_ = X

        # Slight complication to avoid computing knn twice.
        need_knn = 0
        need_knn_ngb = False
        need_knn_dist = False
        if self.graph_type_ == "knn":
            k_graph = self.params_.get("k", 10)
            # If X has fewer than k points...
            if k_graph > len(X):
                k_graph = len(X)
            need_knn = k_graph
            need_knn_ngb = True
        if self.density_type_ in ["DTM", "logDTM"]:
            k = self.params_.get("k", 10)
            k_DTM = self.params_.get("k_DTM", k)
            # If X has fewer than k points...
            if k_DTM > len(X):
                k_DTM = len(X)
            need_knn = max(need_knn, k_DTM)
            need_knn_dist = True
            # if we ask for more neighbors for the graph than the DTM, getting the distances is a slight waste,
            # but it looks negligible
        if need_knn > 0:
            knn_args = dict(self.params_)
            knn_args["k"] = need_knn
            knn = KNearestNeighbors(return_index=need_knn_ngb, return_distance=need_knn_dist, **knn_args).fit_transform(
                X
            )
            if need_knn_ngb:
                if need_knn_dist:
                    self.neighbors_ = knn[0][:, 0:k_graph]
                    knn_dist = knn[1]
                else:
                    self.neighbors_ = knn
            elif need_knn_dist:
                knn_dist = knn
        if self.density_type_ in ["DTM", "logDTM"]:
            dim = self.params_.get("dim")
            if dim is None:
                dim = len(X[0]) if metric != "precomputed" else 2
            q = self.params_.get("q", dim)
            weights = DTMDensity(k=k_DTM, metric="neighbors", dim=dim, q=q).fit_transform(knn_dist)
            if self.density_type_ == "logDTM":
                weights = numpy.log(weights)

        if self.graph_type_ == "radius":
            if metric in ["minkowski", "euclidean", "manhattan", "chebyshev"]:
                from scipy.spatial import cKDTree

                tree = cKDTree(X)
                # TODO: handle "l1" and "l2" aliases?
                p = self.params_.get("p")
                if metric == "euclidean":
                    assert p is None or p == 2, "p=" + str(p) + " is not consistent with metric='euclidean'"
                    p = 2
                elif metric == "manhattan":
                    assert p is None or p == 1, "p=" + str(p) + " is not consistent with metric='manhattan'"
                    p = 1
                elif metric == "chebyshev":
                    assert p is None or p == numpy.inf, "p=" + str(p) + " is not consistent with metric='chebyshev'"
                    p = numpy.inf
                elif p is None:
                    p = 2  # the default
                eps = self.params_.get("eps", 0)
                self.neighbors_ = tree.query_ball_tree(tree, r=self.params_["r"], p=p, eps=eps)

            # TODO: sklearn's NearestNeighbors.radius_neighbors can handle more metrics efficiently via its BallTree
            # (don't bother with the _graph variant, it just calls radius_neighbors).
            elif metric != "precomputed":
                from sklearn.metrics import pairwise_distances

                X = pairwise_distances(X, metric=metric, n_jobs=self.params_.get("n_jobs"))
                metric = "precomputed"

            if metric == "precomputed":
                # TODO: parallelize? May not be worth it.
                X = numpy.asarray(X)
                r = self.params_["r"]
                self.neighbors_ = [numpy.flatnonzero(l <= r) for l in X]

        if self.density_type_ in {"KDE", "logKDE"}:
            # Slow...
            assert (
                self.graph_type_ != "manual" and metric != "precomputed"
            ), "Scikit-learn's KernelDensity requires point coordinates"
            kde_params = dict(self.params_.get("kde_params", dict()))
            kde_params.setdefault("metric", metric)
            r = self.params_.get("r")
            if r is not None:
                kde_params.setdefault("bandwidth", r)
            # Should we default rtol to eps?
            from sklearn.neighbors import KernelDensity

            weights = KernelDensity(**kde_params).fit(self.points_).score_samples(self.points_)
            if self.density_type_ == "KDE":
                # First rescale to avoid computing exp(-1000)
                weights -= numpy.max(weights)
                weights = numpy.exp(weights)

        # TODO: do it at the C++ level and/or in parallel if this is too slow?
        if self.params_.get("symmetrize_graph"):
            self.neighbors_ = [set(line) for line in self.neighbors_]
            for i, line in enumerate(self.neighbors_):
                line.discard(i)
                for j in line:
                    self.neighbors_[j].add(i)

        self.weights_ = weights
        # This is where the main computation happens
        self.leaf_labels_, self.children_, self.diagram_, self.max_weight_per_cc_ = hierarchy(self.neighbors_, weights)
        self.n_leaves_ = len(self.max_weight_per_cc_) + len(self.children_)
        assert self.leaf_labels_.max() + 1 == len(self.max_weight_per_cc_) + len(self.children_)
        # TODO: deduplicate this code with the setters below
        if self.__merge_threshold:
            assert not self.__n_clusters
            self.__n_clusters = numpy.count_nonzero(
                self.diagram_[:, 0] - self.diagram_[:, 1] > self.__merge_threshold
            ) + len(self.max_weight_per_cc_)
        if self.__n_clusters:
            # TODO: set corresponding merge_threshold?
            renaming = merge(self.children_, self.n_leaves_, self.__n_clusters)
            self.labels_ = renaming[self.leaf_labels_]
            # In case the user asked for something impossible.
            # TODO: check for impossible situations before calling merge.
            self.__n_clusters = self.labels_.max() + 1
        else:
            self.labels_ = self.leaf_labels_
            self.__n_clusters = self.n_leaves_
        return self

    def fit_predict(self, X, y=None, weights=None):
        """
        Equivalent to fit(), and returns the `labels_`.
        """
        return self.fit(X, y, weights).labels_

    # TODO: add argument k or threshold? Have a version where you can click and it shows the line and the corresponding k?
    def plot_diagram(self):
        """
        """
        import matplotlib.pyplot as plt

        l = self.max_weight_per_cc_.min()
        r = self.max_weight_per_cc_.max()
        if self.diagram_.size > 0:
            plt.plot(self.diagram_[:, 0], self.diagram_[:, 1], "o", color="red")
            l = min(l, self.diagram_[:, 1].min())
            r = max(r, self.diagram_[:, 0].max())
        if l == r:
            if l > 0:
                l, r = 0.9 * l, 1.1 * r
            elif l < 0:
                l, r = 1.1 * l, 0.9 * r
            else:
                l, r = -1.0, 1.0
        plt.plot([l, r], [l, r])
        plt.plot(
            self.max_weight_per_cc_, numpy.full(self.max_weight_per_cc_.shape, 1.1 * l - 0.1 * r), "o", color="green"
        )
        plt.show()

    # Use set_params instead?
    @property
    def n_clusters_(self):
        return self.__n_clusters

    @n_clusters_.setter
    def n_clusters_(self, n_clusters):
        if n_clusters == self.__n_clusters:
            return
        self.__n_clusters = n_clusters
        self.__merge_threshold = None
        if hasattr(self, "leaf_labels_"):
            renaming = merge(self.children_, self.n_leaves_, self.__n_clusters)
            self.labels_ = renaming[self.leaf_labels_]
            # In case the user asked for something impossible
            self.__n_clusters = self.labels_.max() + 1

    @property
    def merge_threshold_(self):
        return self.__merge_threshold

    @merge_threshold_.setter
    def merge_threshold_(self, merge_threshold):
        if merge_threshold == self.__merge_threshold:
            return
        if hasattr(self, "leaf_labels_"):
            self.n_clusters_ = numpy.count_nonzero(self.diagram_[:, 0] - self.diagram_[:, 1] > merge_threshold) + len(
                self.max_weight_per_cc_
            )
        else:
            self.__n_clusters = None
        self.__merge_threshold = merge_threshold



#!/usr/bin/env python

import matplotlib.pyplot as plot
import gudhi

""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2016 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

print("#####################################################################")
print("Show barcode persistence example")

persistence = [
    (2, (1.0, float("inf"))),
    (1, (1.4142135623730951, float("inf"))),
    (1, (1.4142135623730951, float("inf"))),
    (0, (0.0, float("inf"))),
    (0, (0.0, 1.0)),
    (0, (0.0, 1.0)),
    (0, (0.0, 1.0)),
]
gudhi.plot_persistence_barcode(persistence)
plot.show()

print("#####################################################################")
print("Show diagram persistence example")

gudhi.plot_persistence_diagram(persistence)
plot.show()

print("#####################################################################")
print("Show diagram persistence example with a confidence band")

gudhi.plot_persistence_diagram(persistence, band=0.2)
plot.show()

print("#####################################################################")
print("Show barcode and diagram persistence side by side example")
fig, axes = plot.subplots(nrows=1, ncols=2)
gudhi.plot_persistence_barcode(persistence, axes = axes[0])
gudhi.plot_persistence_diagram(persistence, axes = axes[1])
fig.suptitle("barcode versus diagram")
plot.show()



from .cubical_layer import CubicalLayer
from .lower_star_simplex_tree_layer import LowerStarSimplexTreeLayer
from .rips_layer import RipsLayer

__all__ = ["LowerStarSimplexTreeLayer", "RipsLayer", "CubicalLayer"]



import os
import sys
import matplotlib.pyplot as plt
import numpy as np
import pytest
import random

from sklearn.cluster import KMeans

# Vectorization
from gudhi.representations import (Landscape, Silhouette, BettiCurve, ComplexPolynomial,\
  TopologicalVector, PersistenceImage, Entropy)

# Preprocessing
from gudhi.representations import (BirthPersistenceTransform, Clamping, DiagramScaler, Padding, ProminentPoints, \
  DiagramSelector)

# Kernel
from gudhi.representations import (PersistenceWeightedGaussianKernel, \
  PersistenceScaleSpaceKernel, SlicedWassersteinDistance,\
  SlicedWassersteinKernel, PersistenceFisherKernel, WassersteinDistance)


def test_representations_examples():
    # Disable graphics for testing purposes
    plt.show = lambda: None
    here = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(here + "/../example")
    import diagram_vectorizations_distances_kernels

    return None


from gudhi.representations.vector_methods import Atol
from gudhi.representations.metrics import *
from gudhi.representations.kernel_methods import *


def _n_diags(n):
    l = []
    for _ in range(n):
        a = np.random.rand(50, 2)
        a[:, 1] += a[:, 0]  # So that y >= x
        l.append(a)
    return l

metrics_dict = { # (class, metric_kwargs, tolerance_pytest_approx)
    "bottleneck": (BottleneckDistance(epsilon=0.00001),
                   dict(e=0.00001),
                   dict(abs=1e-5)),
    "wasserstein": (WassersteinDistance(order=2, internal_p=2, n_jobs=4),
                    dict(order=2, internal_p=2, n_jobs=4),
                    dict(rel=1e-3)),
    "sliced_wasserstein": (SlicedWassersteinDistance(num_directions=100, n_jobs=4),
                           dict(num_directions=100),
                           dict(rel=1e-3)),
    "persistence_fisher": (PersistenceFisherDistance(bandwidth=1., n_jobs=4),
                           dict(bandwidth=1., n_jobs=4),
                           dict(abs=1e-5)),
}


def test_distance_transform_consistency():
    l1 = _n_diags(9)
    l1b = l1.copy()
    for metricName, (metricClass, metricParams, tolerance) in metrics_dict.items():
        d1 = pairwise_persistence_diagram_distances(l1, metric=metricName, **metricParams)
        d2 = metricClass.fit_transform(l1)
        assert d1 == pytest.approx(d2)
        d3 = pairwise_persistence_diagram_distances(l1, l1b, metric=metricName, **metricParams)
        assert d3 == pytest.approx(d2, **tolerance)  # Because of 0 entries (on the diagonal)
        d4 = metricClass.fit(l1).transform(l1b)
        assert d4 == pytest.approx(d2, **tolerance)


kernel_dict = {
    "sliced_wasserstein": (SlicedWassersteinKernel(num_directions=10, bandwidth=4., n_jobs=4),
                           dict(num_directions=10), dict(rel=1e-3)),
    "persistence_fisher": (PersistenceFisherKernel(bandwidth_fisher=3., bandwidth=1.),
                           dict(bandwidth=3.),  # corresponds to bandwidth_fisher in the kernel class
                           dict(rel=1e-3)),
    "persistence_weighted_gaussian": (PersistenceWeightedGaussianKernel(bandwidth=4.,
                                                                        weight=lambda x: x[1]-x[0]),
                                      dict(bandwidth=4., weight=lambda x: x[1]-x[0]),
                                      dict(rel=1e-3)),
    "persistence_scale_space": (PersistenceScaleSpaceKernel(bandwidth=4.),
                                      dict(bandwidth=4.),
                                      dict(rel=1e-3)),
}
def test_kernel_from_distance():
    l1, l2 = _n_diags(9), _n_diags(11)
    for kernelName in ["sliced_wasserstein", "persistence_fisher"]:
        kernelClass, kernelParams, tolerance = kernel_dict[kernelName]
        f1 = kernelClass.fit_transform(l1)
        d1 = pairwise_persistence_diagram_distances(l1, metric=kernelName, **kernelParams)
        assert np.exp(-d1/kernelClass.bandwidth == pytest.approx(f1, **tolerance))

def test_kernel_distance_consistency():
    l1, l2 = _n_diags(9), _n_diags(11)
    for kernelName, (kernelClass, kernelParams, tolerance) in kernel_dict.items():
        _ = kernelClass.fit(l1)
        f2 = kernelClass.transform(l2)
        f12 = np.array([[kernelClass(l1_, l2_) for l1_ in l1] for l2_ in l2])
        assert f12 == pytest.approx(f2, **tolerance)

def test_sliced_wasserstein_distance_value():
    diag1 = np.array([[0., 1.], [0., 2.]])
    diag2 = np.array([[1., 0.]])
    SWD = SlicedWassersteinDistance(num_directions=2)
    distance = SWD(diag1, diag2)
    assert distance == pytest.approx(2., abs=1e-8)


# Test sorted values as points order can be inverted, and sorted test is not documentation-friendly
# Note the test below must be up to date with the Atol class documentation
def test_atol_doc():
    a = np.array([[1, 2, 4], [1, 4, 0], [1, 0, 4]])
    b = np.array([[4, 2, 0], [4, 4, 0], [4, 0, 2]])
    c = np.array([[3, 2, -1], [1, 2, -1]])

    atol_vectoriser = Atol(quantiser=KMeans(n_clusters=2, random_state=202006, n_init=10))
    # Atol will do
    # X = np.concatenate([a,b,c])
    # kmeans = KMeans(n_clusters=2, random_state=202006, n_init=10).fit(X)
    # kmeans.labels_ will be : array([1, 0, 1, 0, 0, 1, 0, 0])
    first_cluster = np.asarray([a[0], a[2], b[2]])
    second_cluster = np.asarray([a[1], b[0], b[1], c[0], c[1]])

    # Check the center of the first_cluster and second_cluster are in Atol centers
    centers = atol_vectoriser.fit(X=[a, b, c]).centers
    np.isclose(centers, first_cluster.mean(axis=0)).all(1).any() 
    np.isclose(centers, second_cluster.mean(axis=0)).all(1).any() 

    vectorization = atol_vectoriser.transform(X=[a, b, c])
    assert np.allclose(vectorization[0], atol_vectoriser(a))
    assert np.allclose(vectorization[1], atol_vectoriser(b))
    assert np.allclose(vectorization[2], atol_vectoriser(c))


def test_dummy_atol():
    a = np.array([[1, 2, 4], [1, 4, 0], [1, 0, 4]])
    b = np.array([[4, 2, 0], [4, 4, 0], [4, 0, 2]])
    c = np.array([[3, 2, -1], [1, 2, -1]])

    for weighting_method in ["cloud", "iidproba"]:
        for contrast in ["gaussian", "laplacian", "indicator"]:
            atol_vectoriser = Atol(
                quantiser=KMeans(n_clusters=1, random_state=202006, n_init=10),
                weighting_method=weighting_method,
                contrast=contrast,
            )
            atol_vectoriser.fit([a, b, c])
            atol_vectoriser(a)
            atol_vectoriser.transform(X=[a, b, c])


from gudhi.representations.vector_methods import BettiCurve

def test_infinity():
    a = np.array([[1.0, 8.0], [2.0, np.inf], [3.0, 4.0]])
    c = BettiCurve(20, [0.0, 10.0])(a)
    assert c[1] == 0
    assert c[7] == 3
    assert c[9] == 2

def test_preprocessing_empty_diagrams():
    empty_diag = np.empty(shape = [0, 2])
    assert not np.any(BirthPersistenceTransform()(empty_diag))
    assert not np.any(Clamping().fit_transform(empty_diag))
    assert not np.any(DiagramScaler()(empty_diag))
    assert not np.any(Padding()(empty_diag))
    assert not np.any(ProminentPoints()(empty_diag))
    assert not np.any(DiagramSelector()(empty_diag))

def pow(n):
  return lambda x: np.power(x[1]-x[0],n)

def test_vectorization_empty_diagrams():
    empty_diag = np.empty(shape = [0, 2])
    random_resolution = random.randint(50,100)*10 # between 500 and 1000
    print("resolution = ", random_resolution)
    lsc = Landscape(resolution=random_resolution)(empty_diag)
    assert not np.any(lsc)
    assert lsc.shape[0]%random_resolution == 0
    slt = Silhouette(resolution=random_resolution, weight=pow(2))(empty_diag)
    assert not np.any(slt)
    assert slt.shape[0] == random_resolution
    btc = BettiCurve(resolution=random_resolution)(empty_diag)
    assert not np.any(btc)
    assert btc.shape[0] == random_resolution
    cpp = ComplexPolynomial(threshold=random_resolution, polynomial_type="T")(empty_diag)
    assert not np.any(cpp)
    assert cpp.shape[0] == random_resolution
    tpv = TopologicalVector(threshold=random_resolution)(empty_diag)
    assert tpv.shape[0] == random_resolution
    assert not np.any(tpv)
    prmg = PersistenceImage(resolution=[random_resolution,random_resolution])(empty_diag)
    assert not np.any(prmg)
    assert prmg.shape[0] == random_resolution * random_resolution
    sce = Entropy(mode="scalar", resolution=random_resolution)(empty_diag)
    assert not np.any(sce)
    assert sce.shape[0] == 1
    scv = Entropy(mode="vector", normalized=False, resolution=random_resolution)(empty_diag)
    assert not np.any(scv)
    assert scv.shape[0] == random_resolution
    
def test_entropy_miscalculation():
    diag_ex = np.array([[0.0,1.0], [0.0,1.0], [0.0,2.0]])
    def pe(pd):
        l = pd[:,1] - pd[:,0]
        l = l/sum(l)
        return -np.dot(l, np.log(l))
    sce = Entropy(mode="scalar")
    assert [[pe(diag_ex)]] == sce.fit_transform([diag_ex])
    sce = Entropy(mode="vector", resolution=4, normalized=False, keep_endpoints=True)
    pef = [-1/4*np.log(1/4)-1/4*np.log(1/4)-1/2*np.log(1/2),
           -1/4*np.log(1/4)-1/4*np.log(1/4)-1/2*np.log(1/2),
           -1/2*np.log(1/2), 
           0.0]
    assert all(([pef] == sce.fit_transform([diag_ex]))[0])
    sce = Entropy(mode="vector", resolution=4, normalized=True)
    pefN = (sce.fit_transform([diag_ex]))[0]
    area = np.linalg.norm(pefN, ord=1)
    assert area==pytest.approx(1)
        
def test_kernel_empty_diagrams():
    empty_diag = np.empty(shape = [0, 2])
    assert SlicedWassersteinDistance(num_directions=100)(empty_diag, empty_diag) == 0.
    assert SlicedWassersteinKernel(num_directions=100, bandwidth=1.)(empty_diag, empty_diag) == 1.
    assert WassersteinDistance(mode="hera", delta=0.0001)(empty_diag, empty_diag) == 0.
    assert WassersteinDistance(mode="pot")(empty_diag, empty_diag) == 0.
    assert BottleneckDistance(epsilon=.001)(empty_diag, empty_diag) == 0.
    assert BottleneckDistance()(empty_diag, empty_diag) == 0.
#    PersistenceWeightedGaussianKernel(bandwidth=1., kernel_approx=None, weight=arctan(1.,1.))(empty_diag, empty_diag)
#    PersistenceWeightedGaussianKernel(kernel_approx=RBFSampler(gamma=1./2, n_components=100000).fit(np.ones([1,2])), weight=arctan(1.,1.))(empty_diag, empty_diag)
#    PersistenceScaleSpaceKernel(bandwidth=1.)(empty_diag, empty_diag)
#    PersistenceScaleSpaceKernel(kernel_approx=RBFSampler(gamma=1./2, n_components=100000).fit(np.ones([1,2])))(empty_diag, empty_diag)
#    PersistenceFisherKernel(bandwidth_fisher=1., bandwidth=1.)(empty_diag, empty_diag)
#    PersistenceFisherKernel(bandwidth_fisher=1., bandwidth=1., kernel_approx=RBFSampler(gamma=1./2, n_components=100000).fit(np.ones([1,2])))(empty_diag, empty_diag)


def test_silhouette_permutation_invariance():
    dgm = _n_diags(1)[0]
    dgm_permuted = dgm[np.random.permutation(dgm.shape[0]).astype(int)]
    random_resolution = random.randint(50, 100) * 10
    slt = Silhouette(resolution=random_resolution, weight=pow(2))

    assert np.all(np.isclose(slt(dgm), slt(dgm_permuted)))


def test_silhouette_multiplication_invariance():
    dgm = _n_diags(1)[0]
    n_repetitions = np.random.randint(2, high=10)
    dgm_augmented = np.repeat(dgm, repeats=n_repetitions, axis=0)

    random_resolution = random.randint(50, 100) * 10
    slt = Silhouette(resolution=random_resolution, weight=pow(2))
    assert np.all(np.isclose(slt(dgm), slt(dgm_augmented)))


def test_silhouette_numeric():
    dgm = np.array([[2., 3.], [5., 6.]])
    slt = Silhouette(resolution=9, weight=pow(1), sample_range=[2., 6.])
    #slt.fit([dgm])
    # x_values = array([2., 2.5, 3., 3.5, 4., 4.5, 5., 5.5, 6.])

    expected_silhouette = np.array([0., 0.5, 0., 0., 0., 0., 0., 0.5, 0.])/np.sqrt(2)
    output_silhouette = slt(dgm)
    assert np.all(np.isclose(output_silhouette, expected_silhouette))


def test_landscape_small_persistence_invariance():
    dgm = np.array([[2., 6.], [2., 5.], [3., 7.]])
    small_persistence_pts = np.random.rand(10, 2)
    small_persistence_pts[:, 1] += small_persistence_pts[:, 0]
    small_persistence_pts += np.min(dgm)
    dgm_augmented = np.concatenate([dgm, small_persistence_pts], axis=0)

    lds = Landscape(num_landscapes=2, resolution=5)
    lds_dgm, lds_dgm_augmented = lds(dgm), lds(dgm_augmented)

    assert np.all(np.isclose(lds_dgm, lds_dgm_augmented))


def test_landscape_numeric():
    dgm = np.array([[2., 6.], [3., 5.]])
    lds_ref = np.array([
        0., 0.5, 1., 1.5, 2., 1.5, 1., 0.5, 0.,  # tent of [2, 6]
        0., 0., 0., 0.5, 1., 0.5, 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0.,
        0., 0., 0., 0., 0., 0., 0., 0., 0.,
    ])
    lds_ref *= np.sqrt(2)
    lds = Landscape(num_landscapes=4, resolution=9, sample_range=[2., 6.])
    lds_dgm = lds(dgm)
    assert np.all(np.isclose(lds_dgm, lds_ref))


def test_landscape_nan_range():
    dgm = np.array([[2., 6.], [3., 5.]])
    lds = Landscape(num_landscapes=2, resolution=9, sample_range=[np.nan, 6.])
    lds_dgm = lds.fit([dgm])
    assert lds.sample_range_fixed_[0] == 2 and lds.sample_range_fixed_[1] == 6
    assert lds.new_resolution_ == 10

def test_endpoints():
    diags = [ np.array([[2., 3.]]) ]
    for vec in [ Landscape(), Silhouette(), BettiCurve(), Entropy(mode="vector") ]:
        vec.fit(diags)
        assert vec.grid_[0] > 2 and vec.grid_[-1] < 3
    for vec in [ Landscape(keep_endpoints=True), Silhouette(keep_endpoints=True), BettiCurve(keep_endpoints=True), Entropy(mode="vector", keep_endpoints=True)]:
        vec.fit(diags)
        assert vec.grid_[0] == 2 and vec.grid_[-1] == 3
    vec = BettiCurve(resolution=None)
    vec.fit(diags)
    assert np.equal(vec.grid_, [-np.inf, 2., 3.]).all()

def test_get_params():
    for vec in [ Landscape(), Silhouette(), BettiCurve(), Entropy(mode="vector") ]:
        vec.get_params()



#!/usr/bin/env python

from gudhi import StrongWitnessComplex, SimplexTree

""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2016 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

print("#####################################################################")
print("WitnessComplex creation from nearest landmark table")
nearest_landmark_table = [
    [[0, 0.0], [1, 0.1], [2, 0.2], [3, 0.3], [4, 0.4]],
    [[1, 0.0], [2, 0.1], [3, 0.2], [4, 0.3], [0, 0.4]],
    [[2, 0.0], [3, 0.1], [4, 0.2], [0, 0.3], [1, 0.4]],
    [[3, 0.0], [4, 0.1], [0, 0.2], [1, 0.3], [2, 0.4]],
    [[4, 0.0], [0, 0.1], [1, 0.2], [2, 0.3], [3, 0.4]],
]

witness_complex = StrongWitnessComplex(nearest_landmark_table=nearest_landmark_table)
simplex_tree = witness_complex.create_simplex_tree(max_alpha_square=0.41)

message = "Number of simplices: " + repr(simplex_tree.num_simplices())
print(message)

diag = simplex_tree.persistence(min_persistence=-0.1, homology_coeff_field=11)
print(diag)



from gudhi.tensorflow import *
import numpy as np
import tensorflow as tf
import gudhi as gd

def test_rips_diff():

    Xinit = np.array([[1.,1.],[2.,2.]], dtype=np.float32)
    X = tf.Variable(initial_value=Xinit, trainable=True)
    rl = RipsLayer(maximum_edge_length=2., homology_dimensions=[0])

    with tf.GradientTape() as tape:
        dgm = rl.call(X)[0][0]
        loss = tf.math.reduce_sum(tf.square(.5*(dgm[:,1]-dgm[:,0])))
    grads = tape.gradient(loss, [X])
    assert tf.norm(grads[0]-tf.constant([[-.5,-.5],[.5,.5]]),1) <= 1e-6

def test_cubical_diff():

    Xinit = np.array([[0.,2.,2.],[2.,2.,2.],[2.,2.,1.]], dtype=np.float32)
    X = tf.Variable(initial_value=Xinit, trainable=True)
    cl = CubicalLayer(homology_dimensions=[0])

    with tf.GradientTape() as tape:
        dgm = cl.call(X)[0][0]
        loss = tf.math.reduce_sum(tf.square(.5*(dgm[:,1]-dgm[:,0])))
    grads = tape.gradient(loss, [X])
    assert tf.norm(grads[0]-tf.constant([[0.,0.,0.],[0.,.5,0.],[0.,0.,-.5]]),1) <= 1e-6

def test_nonsquare_cubical_diff():

    Xinit = np.array([[-1.,1.,0.],[1.,1.,1.]], dtype=np.float32)
    X = tf.Variable(initial_value=Xinit, trainable=True)
    cl = CubicalLayer(homology_dimensions=[0])

    with tf.GradientTape() as tape:
        dgm = cl.call(X)[0][0]
        loss = tf.math.reduce_sum(tf.square(.5*(dgm[:,1]-dgm[:,0])))
    grads = tape.gradient(loss, [X])
    assert tf.norm(grads[0]-tf.constant([[0.,0.5,-0.5],[0.,0.,0.]]),1) <= 1e-6

def test_st_diff():

    st = gd.SimplexTree()
    st.insert([0])
    st.insert([1]) 
    st.insert([2]) 
    st.insert([3]) 
    st.insert([4]) 
    st.insert([5]) 
    st.insert([6]) 
    st.insert([7]) 
    st.insert([8]) 
    st.insert([9]) 
    st.insert([10]) 
    st.insert([0, 1]) 
    st.insert([1, 2]) 
    st.insert([2, 3]) 
    st.insert([3, 4]) 
    st.insert([4, 5]) 
    st.insert([5, 6]) 
    st.insert([6, 7]) 
    st.insert([7, 8]) 
    st.insert([8, 9]) 
    st.insert([9, 10]) 

    Finit = np.array([6.,4.,3.,4.,5.,4.,3.,2.,3.,4.,5.], dtype=np.float32)
    F = tf.Variable(initial_value=Finit, trainable=True)
    sl = LowerStarSimplexTreeLayer(simplextree=st, homology_dimensions=[0])

    with tf.GradientTape() as tape:
        dgm = sl.call(F)[0][0]
        loss = tf.math.reduce_sum(tf.square(.5*(dgm[:,1]-dgm[:,0])))
    grads = tape.gradient(loss, [F])

    assert tf.math.reduce_all(tf.math.equal(grads[0].indices, tf.constant([2,4])))
    assert tf.math.reduce_all(tf.math.equal(grads[0].values, tf.constant([-1.,1.])))




# Standard data science imports
import numpy as np
from scipy.stats import bootstrap
from sklearn.pipeline import Pipeline
import matplotlib.pyplot as plt

# Import TDA pipeline requirements
from gudhi.sklearn.rips_persistence import RipsPersistence
from gudhi.representations import DiagramSelector, Landscape
# To fetch the dataset
from gudhi.datasets import remote

no_plot = False

# Mainly for tests purpose - removed from documentation
import argparse
parser = argparse.ArgumentParser(description='Plot average landscapes')
parser.add_argument("--no-plot", default=False, action="store_true")
args = parser.parse_args()
no_plot = args.no_plot

# Constants for subsample
nb_times = 80
nb_points = 200

def subsample(array):
    sub = []
    # construct a list of nb_times x nb_points
    for sub_idx in range(nb_times):
        sub.append(array[np.random.choice(array.shape[0], nb_points, replace=False)])
    return sub

# Constant for plot_average_landscape
landscape_resolution = 600
# Nothing interesting after 0.4, you can set this filter to 1 (number of landscapes) if you want to check
filter = int(0.4 * landscape_resolution)

def plot_average_landscape(landscapes, color, label):
    landscapes = landscapes[:,:filter]
    rng = np.random.default_rng()
    res = bootstrap((np.transpose(landscapes),), np.std, method='basic', axis=-1, confidence_level=0.95, random_state=rng)
    ci_l, ci_u = res.confidence_interval
    plt.fill_between(np.arange(0,filter,1), ci_l, ci_u, alpha=.3, color=color, label=label)

# Fetch datasets and subsample them
walking = subsample(remote.fetch_daily_activities(subset="walking"))
stepper = subsample(remote.fetch_daily_activities(subset="stepper"))
cross = subsample(remote.fetch_daily_activities(subset="cross_training"))
jumping = subsample(remote.fetch_daily_activities(subset="jumping"))

pipe = Pipeline(
    [
        ("rips_pers", RipsPersistence(homology_dimensions=1, n_jobs=-2)),
        ("finite_diags", DiagramSelector(use=True, point_type="finite")),
        ("landscape", Landscape(num_landscapes=1,resolution=landscape_resolution)),
    ]
)

# Fit the model with the complete dataset - mainly for the landscape window computation
pipe.fit(walking + stepper + cross + jumping)

# Compute Rips, persistence, landscapes and plot average landscapes
plot_average_landscape(pipe.transform(walking), 'black', 'walking')
plot_average_landscape(pipe.transform(stepper), 'green', 'stepper')
plot_average_landscape(pipe.transform(cross), 'red', 'cross tr.')
plot_average_landscape(pipe.transform(jumping), 'blue', 'jumping')

plt.title('Average landscapes')
plt.legend()
if no_plot == False:
    plt.show()


# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Raphaël Tinarrage, Yuichi Ike, Masatoshi Takenouchi
#
# Copyright (C) 2020 Inria, Copyright (C) 2020 FUjitsu Laboratories Ltd.
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from gudhi import SimplexTree

class WeightedRipsComplex:
    """
    Class to generate a weighted Rips complex from a distance matrix and weights on vertices, 
    in the way described in :cite:`dtmfiltrations` with `p=1`. The filtration value of vertex `i` is `2*weights[i]`,
    and the filtration value of edge `ij` is `distance_matrix[i][j]+weights[i]+weights[j]`,
    or the maximum of the filtrations of its extremities, whichever is largest.
    Remark that all the filtration values are doubled compared to the definition in the paper 
    for consistency with RipsComplex.
    """
    def __init__(self, 
                distance_matrix, 
                weights=None,
                max_filtration=float('inf')):
        """
        Args:
            distance_matrix (Sequence[Sequence[float]]): distance matrix (full square or lower triangular).
            weights (Sequence[float]): (one half of) weight for each vertex.
            max_filtration (float): specifies the maximal filtration value to be considered.      
        """
        self.distance_matrix = distance_matrix
        if weights is not None:
            self.weights = weights
        else:
            self.weights = [0] * len(distance_matrix)
        self.max_filtration = max_filtration
            
    def create_simplex_tree(self, max_dimension):
        """
        Args:
            max_dimension (int): graph expansion until this given dimension.
        """
        dist = self.distance_matrix
        F = self.weights
        num_pts = len(dist)
        
        st = SimplexTree()
        
        for i in range(num_pts):
            if 2*F[i] <= self.max_filtration:
                st.insert([i], 2*F[i])
        for i in range(num_pts):
            for j in range(i):
                value = max(2*F[i], 2*F[j], dist[i][j] + F[i] + F[j])
                # max is needed when F is not 1-Lipschitz
                if value <= self.max_filtration:
                    st.insert([i,j], filtration=value)
                    
        st.expansion(max_dimension) 
        return st
        



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Vincent Rouvreau, Bertrand Michel
#
# Copyright (C) 2016 Inria
#
# Modification(s):
#   - 2020/02 Theo Lacombe: Added more options for improved rendering and more flexibility.
#   - 2022/11 Vincent Rouvreau: "Automatic" legend display detected by _array_handler that returns if the persistence
#                               was a nx2 array.
#   - YYYY/MM Author: Description of the modification

from os import path
import numpy as np
from functools import lru_cache
import warnings
import errno
import os
import shutil

from gudhi.reader_utils import read_persistence_intervals_in_dimension
from gudhi.reader_utils import read_persistence_intervals_grouped_by_dimension

__author__ = "Vincent Rouvreau, Bertrand Michel, Theo Lacombe"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

_gudhi_matplotlib_use_tex = True


def _min_birth_max_death(persistence, band=0.0):
    """This function returns (min_birth, max_death) from the persistence.

    :param persistence: The persistence to plot.
    :type persistence: list of tuples(dimension, tuple(birth, death)).
    :param band: band
    :type band: float.
    :returns: (float, float) -- (min_birth, max_death).
    """
    # Look for minimum birth date and maximum death date for plot optimisation
    max_death = 0
    min_birth = persistence[0][1][0]
    for interval in reversed(persistence):
        if float(interval[1][1]) != float("inf"):
            if float(interval[1][1]) > max_death:
                max_death = float(interval[1][1])
        if float(interval[1][0]) > max_death:
            max_death = float(interval[1][0])
        if float(interval[1][0]) < min_birth:
            min_birth = float(interval[1][0])
    if band > 0.0:
        max_death += band
    # can happen if only points at inf death
    if min_birth == max_death:
        max_death = max_death + 1.0
    return (min_birth, max_death)


def _array_handler(a):
    """
    :param a: if array, assumes it is a (n x 2) np.array and returns a
                persistence-compatible list (padding with 0), so that the
                plot can be performed seamlessly.
    :returns: * List[dimension, [birth, death]] Persistence, compatible with plot functions, list.
              * boolean Modification status (True if output is different from input)
    """
    if isinstance(a[0][1], (np.floating, float)):
        return [[0, x] for x in a], True
    else:
        return a, False


def _limit_to_max_intervals(persistence, max_intervals, key):
    """This function returns truncated persistence if length is bigger than max_intervals.
    :param persistence: Persistence intervals values list. Can be grouped by dimension or not.
    :type persistence: an array of (dimension, (birth, death)) or an array of (birth, death).
    :param max_intervals: maximal number of intervals to display.
        Selected intervals are those with the longest life time. Set it
        to 0 to see all. Default value is 1000.
    :type max_intervals: int.
    :param key: key function for sort algorithm.
    :type key: function or lambda.
    """
    if max_intervals > 0 and max_intervals < len(persistence):
        warnings.warn(
            "There are %s intervals given as input, whereas max_intervals is set to %s."
            % (len(persistence), max_intervals)
        )
        # Sort by life time, then takes only the max_intervals elements
        return sorted(persistence, key=key, reverse=True)[:max_intervals]
    else:
        return persistence


@lru_cache(maxsize=1)
def _matplotlib_can_use_tex() -> bool:
    """This function returns True if matplotlib can deal with LaTeX, False otherwise.
    The returned value is cached.

    This code is taken
    https://github.com/matplotlib/matplotlib/blob/f975291a008f001047ad8964b15d7d64d2907f1e/lib/matplotlib/__init__.py#L454-L471
    deprecated from matplotlib 3.6 and removed in matplotlib 3.8.0
    """
    from matplotlib import _get_executable_info, ExecutableNotFoundError

    if not shutil.which("tex"):
        warnings.warn("usetex mode requires TeX.")
        return False
    try:
        _get_executable_info("dvipng")
    except ExecutableNotFoundError:
        warnings.warn("usetex mode requires dvipng.")
        return False
    try:
        _get_executable_info("gs")
    except ExecutableNotFoundError:
        warnings.warn("usetex mode requires ghostscript.")
        return False
    return True


def plot_persistence_barcode(
    persistence=[],
    persistence_file="",
    alpha=0.6,
    max_intervals=20000,
    inf_delta=0.1,
    legend=None,
    colormap=None,
    axes=None,
    fontsize=16,
):
    """This function plots the persistence bar code from persistence values list
    , a np.array of shape (N x 2) (representing a diagram
    in a single homology dimension),
    or from a `persistence diagram <fileformats.html#persistence-diagram>`_ file.

    :param persistence: Persistence intervals values list. Can be grouped by dimension or not.
    :type persistence: an array of (dimension, (birth, death)) or an array of (birth, death)
    :param persistence_file: A `persistence diagram <fileformats.html#persistence-diagram>`_ file style name
        (reset persistence if both are set).
    :type persistence_file: string
    :param alpha: barcode transparency value (0.0 transparent through 1.0
        opaque - default is 0.6).
    :type alpha: float
    :param max_intervals: maximal number of intervals to display.
        Selected intervals are those with the longest life time. Set it
        to 0 to see all. Default value is 20000.
    :type max_intervals: int
    :param inf_delta: Infinity is placed at :code:`((max_death - min_birth) x
        inf_delta)` above :code:`max_death` value. A reasonable value is
        between 0.05 and 0.5 - default is 0.1.
    :type inf_delta: float
    :param legend: Display the dimension color legend. Default is None, meaning the legend is displayed if dimension
        is specified in the persistence argument, and not displayed if dimension is not specified.
    :type legend: boolean or None
    :param colormap: A matplotlib-like qualitative colormaps. Default is None
        which means :code:`matplotlib.cm.Set1.colors`.
    :type colormap: tuple of colors (3-tuple of float between 0. and 1.)
    :param axes: A matplotlib-like subplot axes. If None, the plot is drawn on
        a new set of axes.
    :type axes: `matplotlib.axes.Axes`
    :param fontsize: Fontsize to use in axis.
    :type fontsize: int
    :returns: (`matplotlib.axes.Axes`): The axes on which the plot was drawn.
    """
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches

    if _gudhi_matplotlib_use_tex and _matplotlib_can_use_tex():
        plt.rc("text", usetex=True)
        plt.rc("font", family="serif")
    else:
        plt.rc("text", usetex=False)
        plt.rc("font", family="DejaVu Sans")

    # By default, let's say the persistence is not an array of shape (N x 2) - Can be from a persistence file
    nx2_array = False
    if persistence_file != "":
        if path.isfile(persistence_file):
            # Reset persistence
            persistence = []
            diag = read_persistence_intervals_grouped_by_dimension(persistence_file=persistence_file)
            for key in diag.keys():
                for persistence_interval in diag[key]:
                    persistence.append((key, persistence_interval))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), persistence_file)

    try:
        persistence, nx2_array = _array_handler(persistence)
        persistence = _limit_to_max_intervals(
            persistence, max_intervals, key=lambda life_time: life_time[1][1] - life_time[1][0]
        )
        (min_birth, max_death) = _min_birth_max_death(persistence)
        persistence = sorted(persistence, key=lambda birth: birth[1][0])
    except IndexError:
        min_birth, max_death = 0.0, 1.0
        pass

    delta = (max_death - min_birth) * inf_delta
    # Replace infinity values with max_death + delta for bar code to be more
    # readable
    infinity = max_death + delta
    axis_start = min_birth - delta

    if axes is None:
        _, axes = plt.subplots(1, 1)
    if colormap is None:
        colormap = plt.cm.Set1.colors

    x = [birth for (dim, (birth, death)) in persistence]
    y = [(death - birth) if death != float("inf") else (infinity - birth) for (dim, (birth, death)) in persistence]
    c = [colormap[dim] for (dim, (birth, death)) in persistence]

    axes.barh(range(len(x)), y, left=x, alpha=alpha, color=c, linewidth=0)

    if legend is None and not nx2_array:
        # By default, if persistence is an array of (dimension, (birth, death)), display the legend
        legend = True

    if legend:
        dimensions = {item[0] for item in persistence}
        axes.legend(
            handles=[mpatches.Patch(color=colormap[dim], label=str(dim)) for dim in dimensions],
            loc="best",
        )

    axes.set_title("Persistence barcode", fontsize=fontsize)
    axes.set_yticks([])
    axes.invert_yaxis()

    # Ends plot on infinity value and starts a little bit before min_birth
    if len(x) != 0:
        axes.set_xlim((axis_start, infinity))
    return axes


def plot_persistence_diagram(
    persistence=[],
    persistence_file="",
    alpha=0.6,
    band=0.0,
    max_intervals=1000000,
    inf_delta=0.1,
    legend=None,
    colormap=None,
    axes=None,
    fontsize=16,
    greyblock=True,
):
    r"""This function plots the persistence diagram from persistence values
    list, a np.array of shape (N x 2) representing a diagram in a single
    homology dimension, or from a `persistence diagram <fileformats.html#persistence-diagram>`_ file`.

    :param persistence: Persistence intervals values list. Can be grouped by dimension or not.
    :type persistence: an array of (dimension, (birth, death)) or an array of (birth, death)
    :param persistence_file: A `persistence diagram <fileformats.html#persistence-diagram>`_ file style name
        (reset persistence if both are set).
    :type persistence_file: string
    :param alpha: plot transparency value (0.0 transparent through 1.0
        opaque - default is 0.6).
    :type alpha: float
    :param band: band (not displayed if :math:`\leq` 0. - default is 0.)
    :type band: float
    :param max_intervals: maximal number of intervals to display.
        Selected intervals are those with the longest life time. Set it
        to 0 to see all. Default value is 1000000.
    :type max_intervals: int
    :param inf_delta: Infinity is placed at :code:`((max_death - min_birth) x
        inf_delta)` above :code:`max_death` value. A reasonable value is
        between 0.05 and 0.5 - default is 0.1.
    :type inf_delta: float
    :param legend: Display the dimension color legend. Default is None, meaning the legend is displayed if dimension
        is specified in the persistence argument, and not displayed if dimension is not specified.
    :type legend: boolean or None
    :param colormap: A matplotlib-like qualitative colormaps. Default is None
        which means :code:`matplotlib.cm.Set1.colors`.
    :type colormap: tuple of colors (3-tuple of float between 0. and 1.)
    :param axes: A matplotlib-like subplot axes. If None, the plot is drawn on
        a new set of axes.
    :type axes: `matplotlib.axes.Axes`
    :param fontsize: Fontsize to use in axis.
    :type fontsize: int
    :param greyblock: if we want to plot a grey patch on the lower half plane for nicer rendering. Default True.
    :type greyblock: boolean
    :returns: (`matplotlib.axes.Axes`): The axes on which the plot was drawn.
    """
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches

    if _gudhi_matplotlib_use_tex and _matplotlib_can_use_tex():
        plt.rc("text", usetex=True)
        plt.rc("font", family="serif")
    else:
        plt.rc("text", usetex=False)
        plt.rc("font", family="DejaVu Sans")

    # By default, let's say the persistence is not an array of shape (N x 2) - Can be from a persistence file
    nx2_array = False
    if persistence_file != "":
        if path.isfile(persistence_file):
            # Reset persistence
            persistence = []
            diag = read_persistence_intervals_grouped_by_dimension(persistence_file=persistence_file)
            for key in diag.keys():
                for persistence_interval in diag[key]:
                    persistence.append((key, persistence_interval))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), persistence_file)

    try:
        persistence, nx2_array = _array_handler(persistence)
        persistence = _limit_to_max_intervals(
            persistence, max_intervals, key=lambda life_time: life_time[1][1] - life_time[1][0]
        )
        min_birth, max_death = _min_birth_max_death(persistence, band)
    except IndexError:
        min_birth, max_death = 0.0, 1.0
        pass

    delta = (max_death - min_birth) * inf_delta
    # Replace infinity values with max_death + delta for diagram to be more
    # readable
    infinity = max_death + delta
    axis_end = max_death + delta / 2
    axis_start = min_birth - delta

    if axes is None:
        _, axes = plt.subplots(1, 1)
    if colormap is None:
        colormap = plt.cm.Set1.colors
    # bootstrap band
    if band > 0.0:
        x = np.linspace(axis_start, infinity, 1000)
        axes.fill_between(x, x, x + band, alpha=alpha, facecolor="red")
    # lower diag patch
    if greyblock:
        axes.add_patch(
            mpatches.Polygon(
                [[axis_start, axis_start], [axis_end, axis_start], [axis_end, axis_end]],
                fill=True,
                color="lightgrey",
            )
        )
    # line display of equation : birth = death
    axes.plot([axis_start, axis_end], [axis_start, axis_end], linewidth=1.0, color="k")

    x = [birth for (dim, (birth, death)) in persistence]
    y = [death if death != float("inf") else infinity for (dim, (birth, death)) in persistence]
    c = [colormap[dim] for (dim, (birth, death)) in persistence]

    axes.scatter(x, y, alpha=alpha, color=c)
    if float("inf") in (death for (dim, (birth, death)) in persistence):
        # infinity line and text
        axes.plot([axis_start, axis_end], [infinity, infinity], linewidth=1.0, color="k", alpha=alpha)
        # Infinity label
        yt = axes.get_yticks()
        yt = yt[np.where(yt < axis_end)]  # to avoid plotting ticklabel higher than infinity
        yt = np.append(yt, infinity)
        ytl = ["%.3f" % e for e in yt]  # to avoid float precision error
        ytl[-1] = r"$+\infty$"
        axes.set_yticks(yt)
        axes.set_yticklabels(ytl)

    if legend is None and not nx2_array:
        # By default, if persistence is an array of (dimension, (birth, death)), display the legend
        legend = True

    if legend:
        dimensions = list({item[0] for item in persistence})
        axes.legend(
            handles=[mpatches.Patch(color=colormap[dim], label=str(dim)) for dim in dimensions],
            loc="lower right",
        )

    axes.set_xlabel("Birth", fontsize=fontsize)
    axes.set_ylabel("Death", fontsize=fontsize)
    axes.set_title("Persistence diagram", fontsize=fontsize)
    # Ends plot on infinity value and starts a little bit before min_birth
    axes.axis([axis_start, axis_end, axis_start, infinity + delta / 2])
    return axes


def plot_persistence_density(
    persistence=[],
    persistence_file="",
    nbins=300,
    bw_method=None,
    max_intervals=1000,
    dimension=None,
    cmap=None,
    legend=True,
    axes=None,
    fontsize=16,
    greyblock=False,
):
    """This function plots the persistence density from persistence values list, np.array of shape (N x 2) representing
    a diagram in a single homology dimension, or from a `persistence diagram <fileformats.html#persistence-diagram>`_
    file. Be aware that this function does not distinguish the dimension, it is up to you to select the required one.
    This function also does not handle degenerate data set (scipy correlation matrix inversion can fail).

    :Requires: `SciPy <installation.html#scipy>`_

    :param persistence: Persistence intervals values list. Can be grouped by dimension or not.
    :type persistence: an array of (dimension, (birth, death)) or an array of (birth, death)
    :param persistence_file: A `persistence diagram <fileformats.html#persistence-diagram>`_
        file style name (reset persistence if both are set).
    :type persistence_file: string
    :param nbins: Evaluate a gaussian kde on a regular grid of nbins x nbins over data extents (default is 300)
    :type nbins: int
    :param bw_method: The method used to calculate the estimator bandwidth. This can be 'scott', 'silverman', a scalar
        constant or a callable. If a scalar, this will be used directly as kde.factor. If a callable, it should take a
        gaussian_kde instance as only parameter and return a scalar. If None (default), 'scott' is used. See
        `scipy.stats.gaussian_kde documentation
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html>`_
        for more details.
    :type bw_method: str, scalar or callable, optional
    :param max_intervals: maximal number of points used in the density estimation. Selected intervals are those with
        the longest life time. Set it to 0 to see all. Default value is 1000.
    :type max_intervals: int
    :param dimension: the dimension to be selected in the intervals (default is None to mix all dimensions).
    :type dimension: int
    :param cmap: A matplotlib colormap (default is matplotlib.pyplot.cm.hot_r).
    :type cmap: cf. matplotlib colormap
    :param legend: Display the color bar values (default is True).
    :type legend: boolean
    :param axes: A matplotlib-like subplot axes. If None, the plot is drawn on a new set of axes.
    :type axes: `matplotlib.axes.Axes`
    :param fontsize: Fontsize to use in axis.
    :type fontsize: int
    :param greyblock: if we want to plot a grey patch on the lower half plane for nicer rendering. Default False.
    :type greyblock: boolean
    :returns: (`matplotlib.axes.Axes`): The axes on which the plot was drawn.
    """
    import matplotlib.pyplot as plt
    import matplotlib.patches as mpatches
    from scipy.stats import kde

    if _gudhi_matplotlib_use_tex and _matplotlib_can_use_tex():
        plt.rc("text", usetex=True)
        plt.rc("font", family="serif")
    else:
        plt.rc("text", usetex=False)
        plt.rc("font", family="DejaVu Sans")

    if persistence_file != "":
        if dimension is None:
            # All dimension case
            dimension = -1
        if path.isfile(persistence_file):
            persistence_dim = read_persistence_intervals_in_dimension(
                persistence_file=persistence_file, only_this_dim=dimension
            )
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), persistence_file)

    # default cmap value cannot be done at argument definition level as matplotlib is not yet defined.
    if cmap is None:
        cmap = plt.cm.hot_r
    if axes is None:
        _, axes = plt.subplots(1, 1)

    try:
        # if not read from file but given by an argument
        persistence, _ = _array_handler(persistence)
        persistence_dim = np.array(
            [
                (dim_interval[1][0], dim_interval[1][1])
                for dim_interval in persistence
                if (dim_interval[0] == dimension) or (dimension is None)
            ]
        )
        persistence_dim = persistence_dim[np.isfinite(persistence_dim[:, 1])]
        persistence_dim = np.array(
            _limit_to_max_intervals(
                persistence_dim, max_intervals, key=lambda life_time: life_time[1] - life_time[0]
            )
        )

        # Set as numpy array birth and death (remove undefined values - inf and NaN)
        birth = persistence_dim[:, 0]
        death = persistence_dim[:, 1]
        birth_min = birth.min()
        birth_max = birth.max()
        death_min = death.min()
        death_max = death.max()

        # Evaluate a gaussian kde on a regular grid of nbins x nbins over data extents
        k = kde.gaussian_kde([birth, death], bw_method=bw_method)
        xi, yi = np.mgrid[
            birth_min : birth_max : nbins * 1j,
            death_min : death_max : nbins * 1j,
        ]
        zi = k(np.vstack([xi.ravel(), yi.ravel()]))
        # Make the plot
        img = axes.pcolormesh(xi, yi, zi.reshape(xi.shape), cmap=cmap, shading="auto")
        plot_success = True

    # IndexError on empty diagrams, ValueError on only inf death values
    except (IndexError, ValueError):
        birth_min = 0.0
        birth_max = 1.0
        death_min = 0.0
        death_max = 1.0
        plot_success = False
        pass

    # line display of equation : birth = death
    x = np.linspace(death_min, birth_max, 1000)
    axes.plot(x, x, color="k", linewidth=1.0)

    if greyblock:
        axes.add_patch(
            mpatches.Polygon(
                [[birth_min, birth_min], [death_max, birth_min], [death_max, death_max]],
                fill=True,
                color="lightgrey",
            )
        )

    if plot_success and legend:
        plt.colorbar(img, ax=axes)

    axes.set_xlabel("Birth", fontsize=fontsize)
    axes.set_ylabel("Death", fontsize=fontsize)
    axes.set_title("Persistence density", fontsize=fontsize)

    return axes



""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2023 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

from gudhi.datasets.generators import points
from gudhi.sklearn.rips_persistence import RipsPersistence
import random
import pytest

def test_rips_persistence_of_points_on_a_circle():
    # Let's test with 5 point clouds
    NB_PC = 5
    point_clouds = [points.sphere(n_samples = random.randint(100,150), ambient_dim = 2) for _ in range(NB_PC)]
    
    rips = RipsPersistence(homology_dimensions=[0, 1], n_jobs=-2)
    diags = rips.fit_transform(point_clouds)
    # list of one array as an input means list of one array as an output
    assert len(diags) == NB_PC
    for idx in range(NB_PC):
        # H0 + H1
        assert len(diags[idx]) == 2
        # H0
        assert len(diags[idx][0]) == len(point_clouds[idx])
        # H1
        assert len(diags[idx][1]) == 1

def test_h1_only_rips_persistence_of_points_on_a_circle():
    rips = RipsPersistence(homology_dimensions=1, n_jobs=-2)
    diags = rips.fit_transform([points.sphere(n_samples = 150, ambient_dim = 2)])[0]
    assert len(diags) == 1
    assert 0. < diags[0][0] < 0.6
    assert 1. < diags[0][1] < 2.

def test_invalid_input_type():
    rips = RipsPersistence(homology_dimensions=0, input_type='whatsoever')
    with pytest.raises(ValueError):
        rips.fit_transform([points.sphere(n_samples = 10, ambient_dim = 2)])

def test_distance_matrix_rips_persistence_of_points_on_a_circle():
    try:
        from scipy.spatial.distance import cdist
        pts = points.sphere(n_samples = 150, ambient_dim = 2)
        distance_matrix = cdist(pts, pts)
        rips = RipsPersistence(homology_dimensions=1, input_type='lower distance matrix')
        diags = rips.fit_transform([distance_matrix])[0]
        assert len(diags) == 1
        assert 0. < diags[0][0] < 0.6
        assert 1. < diags[0][1] < 2.
    except ValueError:
        pass



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ -
# which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full
# license details.
# Author(s):       Vincent Rouvreau
#
# Copyright (C) 2018 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from __future__ import print_function
from cython cimport numeric
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.string cimport string
from libcpp cimport bool
import errno
import os
from libc.stdint cimport intptr_t

from gudhi.simplex_tree cimport *
from gudhi.simplex_tree import SimplexTree

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2018 Inria"
__license__ = "GPL v3"

cdef extern from "Nerve_gic_interface.h" namespace "Gudhi":
    cdef cppclass Nerve_gic_interface "Gudhi::cover_complex::Nerve_gic_interface":
        Nerve_gic_interface()
        double compute_confidence_level_from_distance(double distance)
        double compute_distance_from_confidence_level(double alpha)
        void compute_distribution(int N)
        double compute_p_value()
        vector[pair[double, double]] compute_PD()
        void find_simplices()
        void create_simplex_tree(Simplex_tree_python_interface* simplex_tree)
        bool read_point_cloud(string off_file_name)
        double set_automatic_resolution()
        void set_color_from_coordinate(int k)
        void set_color_from_file(string color_file_name)
        void set_color_from_range(vector[double] color)
        void set_cover_from_file(string cover_file_name)
        void set_cover_from_range(vector[vector[int]] assignments)
        void set_cover_from_function()
        void set_cover_from_Euclidean_Voronoi(int m)
        void set_function_from_coordinate(int k)
        void set_function_from_file(string func_file_name)
        void set_function_from_range(vector[double] function)
        void set_gain(double g)
        double set_graph_from_automatic_euclidean_rips(int N)
        void set_graph_from_file(string graph_file_name)
        void set_graph_from_OFF()
        void set_graph_from_euclidean_rips(double threshold)
        void set_mask(int nodemask)
        void set_resolution_with_interval_length(double resolution)
        void set_resolution_with_interval_number(int resolution)
        void set_subsampling(double constant, double power)
        void set_type(string type)
        void set_verbose(bool verbose)
        vector[int] subpopulation(int c)
        double subcolor(int c)
        void write_info()
        void plot_DOT()
        void plot_OFF()
        void set_point_cloud_from_range(vector[vector[double]] cloud)
        void set_distances_from_range(vector[vector[double]] distance_matrix)

# CoverComplex python interface
cdef class CoverComplex:
    """Cover complex data structure.

    The data structure is a simplicial complex, representing a Graph Induced
    simplicial Complex (GIC) or a Nerve, and whose simplices are computed with
    a cover C of a point cloud P, which often comes from the preimages of
    intervals covering the image of a function f defined on P. These intervals
    are parameterized by their resolution (either their length or their number)
    and their gain (percentage of overlap). To compute a GIC, one also needs a
    graph G built on top of P, whose cliques with vertices belonging to
    different elements of C correspond to the simplices of the GIC.
    """

    cdef Nerve_gic_interface * thisptr

    # Fake constructor that does nothing but documenting the constructor
    def __init__(self):
        """CoverComplex constructor.
        """

    # The real cython constructor
    def __cinit__(self):
        self.thisptr = new Nerve_gic_interface()

    def __dealloc__(self):
        if self.thisptr != NULL:
            del self.thisptr

    def _is_defined(self):
        """Returns true if CoverComplex pointer is not NULL.
         """
        return self.thisptr != NULL

    def set_point_cloud_from_range(self, cloud):
        """ Reads and stores the input point cloud from a vector stored in
        memory.

        :param cloud: Input vector containing the point cloud.
        :type cloud: vector[vector[double]]
        """
        return self.thisptr.set_point_cloud_from_range(cloud)

    def set_distances_from_range(self, distance_matrix):
        """ Reads and stores the input distance matrix from a vector stored in
        memory.

        :param distance_matrix: Input vector containing the distance matrix.
        :type distance_matrix: vector[vector[double]]
        """
        return self.thisptr.set_distances_from_range(distance_matrix)

    def compute_confidence_level_from_distance(self, distance):
        """Computes the confidence level of a specific bottleneck distance
        threshold.

        :param distance: Bottleneck distance.
        :type distance: double

        :rtype: double
        :returns: Confidence level.
        """
        return self.thisptr.compute_confidence_level_from_distance(distance)

    def compute_distance_from_confidence_level(self, alpha):
        """Computes the bottleneck distance threshold corresponding to a
        specific confidence level.

        :param alpha: Confidence level.
        :type alpha: double

        :rtype: double
        :returns: Bottleneck distance.
        """
        return self.thisptr.compute_distance_from_confidence_level(alpha)

    def compute_distribution(self, N=100):
        """Computes bootstrapped distances distribution.

        :param N: Loop number (default value is 100).
        :type alpha: int
        """
        self.thisptr.compute_distribution(N)

    def compute_p_value(self):
        """Computes the p-value, i.e. the opposite of the confidence level of
        the largest bottleneck distance preserving the points in the
        persistence diagram of the output simplicial complex.

        :rtype: double
        :returns: p-value.
        """
        return self.thisptr.compute_p_value()

    def compute_PD(self):
        """Computes the extended persistence diagram of the complex.
        """
        return self.thisptr.compute_PD()

    def create_simplex_tree(self):
        """
        :returns: A simplex tree created from the Cover complex.
        :rtype: SimplexTree
        """
        stree = SimplexTree()
        cdef intptr_t stree_int_ptr=stree.thisptr
        self.thisptr.create_simplex_tree(
            <Simplex_tree_python_interface*>stree_int_ptr)
        return stree

    def find_simplices(self):
        """Computes the simplices of the simplicial complex.
        """
        self.thisptr.find_simplices()

    def read_point_cloud(self, off_file):
        """Reads and stores the input point cloud from .(n)OFF file.

        :param off_file: Name of the input .OFF or .nOFF file.
        :type off_file: string

        :rtype: bool
        :returns: Read file status.
        """
        if os.path.isfile(off_file):
            return self.thisptr.read_point_cloud(off_file.encode('utf-8'))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                    off_file)

    def set_automatic_resolution(self):
        """Computes the optimal length of intervals (i.e. the smallest interval
        length avoiding discretization artifacts - see :cite:`Carriere17c`) for a
        functional cover.

        :rtype: double
        :returns: reso interval length used to compute the cover.
        """
        return self.thisptr.set_automatic_resolution()

    def set_color_from_coordinate(self, k=0):
        """Computes the function used to color the nodes of the simplicial
        complex from the k-th coordinate.

        :param k: Coordinate to use (start at 0). Default value is 0.
        :type k: int
        """
        return self.thisptr.set_color_from_coordinate(k)

    def set_color_from_file(self, color_file_name):
        """Computes the function used to color the nodes of the simplicial
        complex from a file containing the function values.

        :param color_file_name: Name of the input color file.
        :type color_file_name: string
        """
        if os.path.isfile(color_file_name):
            self.thisptr.set_color_from_file(color_file_name.encode('utf-8'))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                    color_file_name)

    def set_color_from_range(self, color):
        """Computes the function used to color the nodes of the simplicial
        complex from a vector stored in memory.

        :param color: Input vector of values.
        :type color: vector[double]
        """
        self.thisptr.set_color_from_range(color)

    def set_cover_from_file(self, cover_file_name):
        """Creates the cover C from a file containing the cover elements of
        each point (the order has to be the same as in the input file!).

        :param cover_file_name: Name of the input cover file.
        :type cover_file_name: string
        """
        if os.path.isfile(cover_file_name):
            self.thisptr.set_cover_from_file(cover_file_name.encode('utf-8'))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                    cover_file_name)

    def set_cover_from_range(self, assignments):
        """Creates a cover C from a vector stored in memory.

        :param assignments: Vector containing the assignments of the points to their corresponding cover elements. For instance, if the i-th point belongs to the 1st and 3rd cover elements, then assignments[i] = [1,3].
        :type assignments: List[List[int]]
        """
        self.thisptr.set_cover_from_range(assignments)

    def set_cover_from_function(self):
        """Creates a cover C from the preimages of the function f.
        """
        self.thisptr.set_cover_from_function()

    def set_cover_from_Voronoi(self, m=100):
        """Creates the cover C from the Voronoï cells of a subsampling of the
        point cloud.

        :param m: Number of points in the subsample. Default value is 100.
        :type m: int
        """
        self.thisptr.set_cover_from_Euclidean_Voronoi(m)

    def set_function_from_coordinate(self, k):
        """Creates the function f from the k-th coordinate of the point cloud.

        :param k: Coordinate to use (start at 0).
        :type k: int
        """
        self.thisptr.set_function_from_coordinate(k)

    def set_function_from_file(self, func_file_name):
        """Creates the function f from a file containing the function values.

        :param func_file_name: Name of the input function file.
        :type func_file_name: string
        """
        if os.path.isfile(func_file_name):
            self.thisptr.set_function_from_file(func_file_name.encode('utf-8'))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                    func_file_name)

    def set_function_from_range(self, function):
        """Creates the function f from a vector stored in memory.

        :param function: Input vector of values.
        :type function: vector[double]
        """
        self.thisptr.set_function_from_range(function)

    def set_gain(self, g = 0.3):
        """Sets a gain from a value stored in memory.

        :param g: Gain (default value is 0.3).
        :type g: double
        """
        self.thisptr.set_gain(g)

    def set_graph_from_automatic_rips(self, N=100):
        """Creates a graph G from a Rips complex whose threshold value is
        automatically tuned with subsampling - see :cite:`Carriere17c`.

        :param N: Number of subsampling iteration (the default reasonable value is 100, but there is no guarantee on how to choose it).
        :type N: int
        :rtype: double
        :returns: Delta threshold used for computing the Rips complex.
        """
        return self.thisptr.set_graph_from_automatic_euclidean_rips(N)

    def set_graph_from_file(self, graph_file_name):
        """Creates a graph G from a file containing the edges.

        :param graph_file_name: Name of the input graph file. The graph file
            contains one edge per line, each edge being represented by the IDs
            of its two nodes.
        :type graph_file_name: string
        """
        if os.path.isfile(graph_file_name):
            self.thisptr.set_graph_from_file(graph_file_name.encode('utf-8'))
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                    graph_file_name)

    def set_graph_from_OFF(self):
        """Creates a graph G from the triangulation given by the input OFF
        file.
        """
        self.thisptr.set_graph_from_OFF()

    def set_graph_from_rips(self, threshold):
        """Creates a graph G from a Rips complex.

        :param threshold: Threshold value for the Rips complex.
        :type threshold: double
        """
        self.thisptr.set_graph_from_euclidean_rips(threshold)

    def set_mask(self, nodemask):
        """Sets the mask, which is a threshold integer such that nodes in the
        complex that contain a number of data points which is less than or
        equal to this threshold are not displayed.

        :param nodemask: Threshold.
        :type nodemask: int
        """
        self.thisptr.set_mask(nodemask)

    def set_resolution_with_interval_length(self, resolution):
        """Sets a length of intervals from a value stored in memory.

        :param resolution: Length of intervals.
        :type resolution: double
        """
        self.thisptr.set_resolution_with_interval_length(resolution)

    def set_resolution_with_interval_number(self, resolution):
        """Sets a number of intervals from a value stored in memory.

        :param resolution: Number of intervals.
        :type resolution: int
        """
        self.thisptr.set_resolution_with_interval_number(resolution)

    def set_subsampling(self, constant, power):
        """Sets the constants used to subsample the data set. These constants
        are explained in :cite:`Carriere17c`.

        :param constant: Constant.
        :type constant: double

        :param power: Power.
        :type resolution: double
        """
        self.thisptr.set_subsampling(constant, power)

    def set_type(self, type):
        """Specifies whether the type of the output simplicial complex.

        :param type: either "GIC" or "Nerve".
        :type type: string
        """
        self.thisptr.set_type(type.encode('utf-8'))

    def set_verbose(self, verbose):
        """Specifies whether the program should display information or not.

        :param verbose: true = display info, false = do not display info.
        :type verbose: boolean
        """
        self.thisptr.set_verbose(verbose)

    def subpopulation(self, c):
        """Returns the data subset corresponding to a specific node of the
        created complex.

        :param c: ID of the node.
        :type c: int

        :rtype: vector[int]
        :returns: Vector of IDs of data points.
        """
        return self.thisptr.subpopulation(c)

    def subcolor(self, c):
        """Returns the mean color value corresponding to a specific node of the
        created complex.

        :param c: ID of the node.
        :type c: int

        :rtype: float
        :returns: Mean color value of data points.
        """
        return self.thisptr.subcolor(c)

    def write_info(self):
        """Creates a .txt file called SC.txt describing the 1-skeleton, which can
        then be plotted with e.g. KeplerMapper.
        """
        return self.thisptr.write_info()

    def plot_dot(self):
        """Creates a .dot file called SC.dot for neato (part of the graphviz
        package) once the simplicial complex is computed to get a visualization of
        its 1-skeleton in a .pdf file.
        """
        return self.thisptr.plot_DOT()

    def plot_off(self):
        """Creates a .off file called SC.off for 3D visualization, which contains
        the 2-skeleton of the GIC. This function assumes that the cover has been
        computed with Voronoi. If data points are in 1D or 2D, the remaining
        coordinates of the points embedded in 3D are set to 0.
        """
        return self.thisptr.plot_OFF()



# Standard scientific Python imports
import numpy as np

# Standard scikit-learn imports
from sklearn.datasets import fetch_openml
from sklearn.pipeline import Pipeline
from sklearn.model_selection import train_test_split
from sklearn.svm import SVC
from sklearn import metrics

# Import TDA pipeline requirements
from gudhi.sklearn.cubical_persistence import CubicalPersistence
from gudhi.representations import PersistenceImage, DiagramSelector

X, y = fetch_openml("mnist_784", version=1, return_X_y=True, as_frame=False)

# Take only the 500 images from this 70.000 images dataset
# Remove these 2 lines if you want to test the full dataset
X = X[:500]
y = y[:500]
# Reshapes the 1d inputs as images in 28 x 28.
X = X.reshape((-1, 28, 28))

# Target is: "is an eight ?"
y = (y == "8") * 1
print("There are", np.sum(y), "eights out of", len(y), "numbers.")

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.4, random_state=0)
pipe = Pipeline(
    [
        ("cub_pers", CubicalPersistence(homology_dimensions=0, n_jobs=-2)),
        # Or for multiple persistence dimension computation
        # ("cub_pers", CubicalPersistence(homology_dimensions=[0, 1])),
        # ("H0_diags", DimensionSelector(index=0), # where index is the index in homology_dimensions array
        ("finite_diags", DiagramSelector(use=True, point_type="finite")),
        (
            "pers_img",
            PersistenceImage(bandwidth=50, weight=lambda x: x[1] ** 2, im_range=[0, 256, 0, 256], resolution=[20, 20]),
        ),
        ("svc", SVC()),
    ]
)

# Learn from the train subset
pipe.fit(X_train, y_train)
# Predict from the test subset
predicted = pipe.predict(X_test)

print(f"Classification report for TDA pipeline {pipe}:\n" f"{metrics.classification_report(y_test, predicted)}\n")

report = metrics.classification_report(y_test, predicted, output_dict=True)
assert report["0"]["f1-score"] > 0.65
assert report["1"]["f1-score"] > 0.65



""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Marc Glisse

    Copyright (C) 2022 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

import gudhi as gd
import numpy as np
import pytest
from tempfile import NamedTemporaryFile


def test_off_rw():
    for dim in range(2, 6):
        X = np.random.rand(123, dim)
        gd.write_points_to_off_file("rand.off", X)
        Y = gd.read_points_from_off_file("rand.off")
        assert Y == pytest.approx(X)


def test_human_off():
    pts = gd.read_points_from_off_file("human.off")
    # Should not try to read faces
    assert pts.shape == (4706, 3)


def test_invalid_off_file():
    name = NamedTemporaryFile().name
    with open(name, "w") as f:
        f.write("nOFL\n1000 200 2\n1.2 1.3 1.4")
    with pytest.raises(ValueError):
        gd.read_points_from_off_file(name)
    # Try to open a non-existing file - a new temp file name should not exist
    with pytest.raises(FileNotFoundError):
        gd.read_points_from_off_file(NamedTemporaryFile().name)


def test_off_file_header():
    name = NamedTemporaryFile().name
    print(name)
    # OFF
    ## classic case
    random_nb_points = np.random.randint(0, 100)
    with open(name, "w") as f:
        f.write(f"OFF\n{random_nb_points} 200 2\n1.2 1.3 1.4")
    with open(name) as f:
        assert gd.off_utils._read_off_file_header(f) == (3, random_nb_points)
    ## with comments
    random_nb_points = np.random.randint(0, 100)
    with open(name, "w") as f:
        f.write(
            f"# comment on the first line\nOFF\n# comment on the third line\n{random_nb_points} 200 2\n"
            "# comment before points\n1.2 1.3 1.4"
        )
    with open(name) as f:
        assert gd.off_utils._read_off_file_header(f) == (3, random_nb_points)
    # nOFF
    ## when 'dim nb_vertices nb_faces nb_edges' on the same line
    random_nb_points = np.random.randint(0, 100)
    random_dim = np.random.randint(3, 100)
    with open(name, "w") as f:
        f.write(f"nOFF\n{random_dim} {random_nb_points} 200 2\n1.2 1.3 1.4")
    with open(name) as f:
        assert gd.off_utils._read_off_file_header(f) == (random_dim, random_nb_points)
    ## when 'dim nb_vertices nb_faces nb_edges' on the same line + comments
    random_nb_points = np.random.randint(0, 100)
    random_dim = np.random.randint(3, 100)
    with open(name, "w") as f:
        f.write(
            f"# comment on the first line\nnOFF\n# comment on the third line\n{random_dim} {random_nb_points} 200 2\n"
            "# comment before points\n1.2 1.3 1.4"
        )
    with open(name) as f:
        assert gd.off_utils._read_off_file_header(f) == (random_dim, random_nb_points)
    ## when 'dim' and 'nb_vertices nb_faces nb_edges' on the separated lines
    random_nb_points = np.random.randint(0, 100)
    random_dim = np.random.randint(3, 100)
    with open(name, "w") as f:
        f.write(f"nOFF\n{random_dim}\n{random_nb_points} 200 2\n1.2 1.3 1.4")
    with open(name) as f:
        assert gd.off_utils._read_off_file_header(f) == (random_dim, random_nb_points)
    ## when 'dim' and 'nb_vertices nb_faces nb_edges' on the separated lines + comments
    random_nb_points = np.random.randint(0, 100)
    random_dim = np.random.randint(3, 100)
    with open(name, "w") as f:
        f.write(
            f"# first comment\nnOFF\n# second comment\n{random_dim}\n# third comment\n{random_nb_points} 200 2\n"
            "# another comment\n1.2 1.3 1.4"
        )
    with open(name) as f:
        assert gd.off_utils._read_off_file_header(f) == (random_dim, random_nb_points)

def test_get_next_line():
    for end_of_line in ['\r\n', '\n']:
        for comment_char in ['#', '@']:
            name = NamedTemporaryFile().name
            print(name)
            with open(name, "w") as f:
                f.write(f"{comment_char}{end_of_line}{end_of_line}{comment_char}{end_of_line}First{end_of_line}")
                f.write(f"{end_of_line}{comment_char}{end_of_line}{end_of_line}Second{end_of_line}{end_of_line}")
            with open(name) as f:
                assert gd.off_utils._get_next_line(f, comment=comment_char).split() == ["First"]
                assert gd.off_utils._get_next_line(f, comment=comment_char).split() == ["Second"]



#!/usr/bin/env python

import gudhi
import numpy
from functools import reduce
import argparse
import operator


""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2016 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

parser = argparse.ArgumentParser(
    description="Random cubical complex.",
    epilog="Example: "
    "./random_cubical_complex_persistence_example.py"
    " 10 10 10 - Constructs a random cubical "
    "complex in a dimension [10, 10, 10] (aka. "
    "1000 random top dimensional cells).",
)
parser.add_argument("dimension", type=int, nargs="*", help="Cubical complex dimensions")

args = parser.parse_args()
dimension_multiplication = reduce(operator.mul, args.dimension, 1)

if dimension_multiplication > 1:
    print("#####################################################################")
    print("CubicalComplex creation")
    cubical_complex = gudhi.CubicalComplex(
        dimensions=args.dimension,
        top_dimensional_cells=numpy.random.rand(dimension_multiplication),
    )

    print("persistence(homology_coeff_field=2, min_persistence=0)=")
    print(cubical_complex.persistence(homology_coeff_field=2, min_persistence=0))

    print("betti_numbers()=")
    print(cubical_complex.betti_numbers())



#!/usr/bin/env python

import gudhi
import matplotlib.pyplot as plt
import time

""" This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
    See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
    Author(s):       Vincent Rouvreau

    Copyright (C) 2016 Inria

    Modification(s):
      - YYYY/MM Author: Description of the modification
"""

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2020 Inria"
__license__ = "MIT"


print("#####################################################################")
print("RipsComplex (only the one-skeleton) creation from tore3D_300.off file")

off_file = gudhi.__root_source_dir__ + '/data/points/tore3D_300.off'
point_cloud = gudhi.read_points_from_off_file(off_file = off_file)
rips_complex = gudhi.RipsComplex(points=point_cloud, max_edge_length=12.0)
simplex_tree = rips_complex.create_simplex_tree(max_dimension=1)
print('1. Rips complex is of dimension ', simplex_tree.dimension(), ' - ',
      simplex_tree.num_simplices(), ' simplices - ',
      simplex_tree.num_vertices(), ' vertices.')

# Expansion of this one-skeleton would require a lot of memory. Let's collapse it
start = time.process_time()
simplex_tree.collapse_edges()
print('2. Rips complex is of dimension ', simplex_tree.dimension(), ' - ',
      simplex_tree.num_simplices(), ' simplices - ',
      simplex_tree.num_vertices(), ' vertices.')
simplex_tree.expansion(3)
diag = simplex_tree.persistence()
print("Collapse, expansion and persistence computation took ", time.process_time() - start, " sec.")

# Use subplots to display diagram and density side by side
fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12, 5))
gudhi.plot_persistence_diagram(diag, axes=axes[0])
axes[0].set_title("Persistence after 1 collapse")

# Collapse can be performed several times. Let's collapse it 3 times
start = time.process_time()
simplex_tree.collapse_edges(nb_iterations = 3)
print('3. Rips complex is of dimension ', simplex_tree.dimension(), ' - ',
      simplex_tree.num_simplices(), ' simplices - ',
      simplex_tree.num_vertices(), ' vertices.')
simplex_tree.expansion(3)
diag = simplex_tree.persistence()
print("Collapse, expansion and persistence computation took ", time.process_time() - start, " sec.")

gudhi.plot_persistence_diagram(diag, axes=axes[1])
axes[1].set_title("Persistence after 3 more collapses")

# Plot the 2 persistence diagrams side to side to check the persistence is the same
plt.show()


# This file is part of the Gudhi Library - https://gudhi.inria.fr/ -
# which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full
# license details.
# Author(s):       Vincent Rouvreau
#
# Copyright (C) 2016 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from __future__ import print_function
from cython cimport numeric
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.string cimport string
from libcpp cimport bool
from libc.stdint cimport intptr_t
import warnings

from gudhi.simplex_tree cimport *
from gudhi.simplex_tree import SimplexTree
from gudhi import read_points_from_off_file

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "GPL v3"

cdef extern from "Alpha_complex_interface.h" namespace "Gudhi":
    cdef cppclass Alpha_complex_interface "Gudhi::alpha_complex::Alpha_complex_interface":
        Alpha_complex_interface(vector[vector[double]] points, vector[double] weights, bool fast_version, bool exact_version) nogil except +
        vector[double] get_point(int vertex) nogil except +
        void create_simplex_tree(Simplex_tree_python_interface* simplex_tree, double max_alpha_square, bool default_filtration_value) nogil except +
        @staticmethod
        void set_float_relative_precision(double precision) nogil
        @staticmethod
        double get_float_relative_precision() nogil

# AlphaComplex python interface
cdef class AlphaComplex:
    """AlphaComplex is a simplicial complex constructed from the finite cells of a Delaunay Triangulation.

    The filtration value of each simplex is computed as the square of the circumradius of the simplex if the
    circumsphere is empty (the simplex is then said to be Gabriel), and as the minimum of the filtration values of the
    codimension 1 cofaces that make it not Gabriel otherwise.

    All simplices that have a filtration value strictly greater than a given alpha squared value are not inserted into
    the complex.

    .. note::

        When Alpha_complex is constructed with an infinite value of alpha, the complex is a Delaunay complex.
    """

    cdef Alpha_complex_interface * this_ptr

    # Fake constructor that does nothing but documenting the constructor
    def __init__(self, points=[], off_file='', weights=None, precision='safe'):
        """AlphaComplex constructor.

        :param points: A list of points in d-Dimension.
        :type points: Iterable[Iterable[float]]

        :param off_file: **[deprecated]** An `OFF file style <fileformats.html#off-file-format>`_ name.
            If an `off_file` is given with `points` as arguments, only points from the file are taken into account.
        :type off_file: string

        :param weights: A list of weights. If set, the number of weights must correspond to the number of points.
        :type weights: Iterable[float]

        :param precision: Alpha complex precision can be 'fast', 'safe' or 'exact'. Default is 'safe'.
        :type precision: string

        :raises FileNotFoundError: **[deprecated]** If `off_file` is set but not found.
        :raises ValueError: In case of inconsistency between the number of points and weights.
        """

    # The real cython constructor
    def __cinit__(self, points = [], off_file = '', weights=None, precision = 'safe'):
        assert precision in ['fast', 'safe', 'exact'], "Alpha complex precision can only be 'fast', 'safe' or 'exact'"
        cdef bool fast = precision == 'fast'
        cdef bool exact = precision == 'exact'

        if off_file:
            warnings.warn("off_file is a deprecated parameter, please consider using gudhi.read_points_from_off_file",
                          DeprecationWarning)
            points = read_points_from_off_file(off_file = off_file)

        # weights are set but is inconsistent with the number of points
        if weights is not None and len(weights) != len(points):
            raise ValueError("Inconsistency between the number of points and weights")

        # need to copy the points to use them without the gil
        cdef vector[vector[double]] pts
        cdef vector[double] wgts
        pts = points
        if weights is not None:
            wgts = weights
        with nogil:
            self.this_ptr = new Alpha_complex_interface(pts, wgts, fast, exact)

    def __dealloc__(self):
        if self.this_ptr != NULL:
            del self.this_ptr

    def _is_defined(self):
        """Returns true if AlphaComplex pointer is not NULL.
         """
        return self.this_ptr != NULL

    def get_point(self, vertex):
        """This function returns the point corresponding to a given vertex from the :class:`~gudhi.SimplexTree`.

        :param vertex: The vertex.
        :type vertex: int
        :rtype: list of float
        :returns: the point.
        """
        return self.this_ptr.get_point(vertex)

    def create_simplex_tree(self, max_alpha_square = float('inf'), default_filtration_value = False):
        """
        :param max_alpha_square: The maximum alpha square threshold the simplices shall not exceed. Default is set to
            infinity, and there is very little point using anything else since it does not save time.
        :type max_alpha_square: float
        :param default_filtration_value: Set this value to `True` if filtration values are not needed to be computed
            (will be set to `NaN`). Default value is `False` (which means compute the filtration values).
        :type default_filtration_value: bool
        :returns: A simplex tree created from the Delaunay Triangulation.
        :rtype: SimplexTree
        """
        stree = SimplexTree()
        cdef double mas = max_alpha_square
        cdef intptr_t stree_int_ptr=stree.thisptr
        cdef bool compute_filtration = default_filtration_value == True
        with nogil:
            self.this_ptr.create_simplex_tree(<Simplex_tree_python_interface*>stree_int_ptr,
                                              mas, compute_filtration)
        return stree

    @staticmethod
    def set_float_relative_precision(precision):
        """
        :param precision: When the AlphaComplex is constructed with :code:`precision = 'safe'` (the default),
            one can set the float relative precision of filtration values computed in
            :func:`~gudhi.AlphaComplex.create_simplex_tree`.
            Default is :code:`1e-5` (cf. :func:`~gudhi.AlphaComplex.get_float_relative_precision`).
            For more details, please refer to
            `CGAL::Lazy_exact_nt<NT>::set_relative_precision_of_to_double <https://doc.cgal.org/latest/Number_types/classCGAL_1_1Lazy__exact__nt.html>`_
        :type precision: float
        """
        if precision <=0. or precision >= 1.:
            raise ValueError("Relative precision value must be strictly greater than 0 and strictly lower than 1")
        Alpha_complex_interface.set_float_relative_precision(precision)
    
    @staticmethod
    def get_float_relative_precision():
        """
        :returns: The float relative precision of filtration values computation in
            :func:`~gudhi.AlphaComplex.create_simplex_tree` when the AlphaComplex is constructed with
            :code:`precision = 'safe'` (the default).
        :rtype: float
        """
        return Alpha_complex_interface.get_float_relative_precision()



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Hind Montassif
#
# Copyright (C) 2021 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification
#   - 2022/09 Vincent Rouvreau: factorize _fetch_remote_license
#                               fetch activities dataset

from os.path import join, split, exists, expanduser
from os import makedirs, remove, environ

from urllib.request import urlretrieve
import hashlib
import shutil
from functools import lru_cache

import numpy as np

def _get_data_home(data_home=None):
    """
    Return the path of the remote datasets directory.
    This folder is used to store remotely fetched datasets.
    By default the datasets directory is set to a folder named 'gudhi_data' in the user home folder.
    Alternatively, it can be set by the 'GUDHI_DATA' environment variable.
    The '~' symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : string
        The path to remote datasets directory.
        Default is `None`, meaning that the data home directory will be set to "~/gudhi_data",
        if the 'GUDHI_DATA' environment variable does not exist.

    Returns
    -------
    data_home: string
        The path to remote datasets directory.
    """
    if data_home is None:
        data_home = environ.get("GUDHI_DATA", join("~", "gudhi_data"))
    data_home = expanduser(data_home)
    makedirs(data_home, exist_ok=True)
    return data_home


def clear_data_home(data_home=None):
    """
    Delete the data home cache directory and all its content.

    Parameters
    ----------
    data_home : string, default is None.
        The path to remote datasets directory.
        If `None` and the 'GUDHI_DATA' environment variable does not exist,
        the default directory to be removed is set to "~/gudhi_data".
    """
    # On windows, needs to clear cache before removing a file - no lazy file deletion
    _load_and_cache_activity.cache_clear()
    data_home = _get_data_home(data_home)
    shutil.rmtree(data_home)


def _checksum_sha256(file_path):
    """
    Compute the file checksum using sha256.

    Parameters
    ----------
    file_path: string
        Full path of the created file including filename.

    Returns
    -------
        The hex digest of file_path.
    """
    sha256_hash = hashlib.sha256()
    chunk_size = 4096
    with open(file_path, "rb") as f:
        # Read and update hash string value in blocks of 4K
        while True:
            buffer = f.read(chunk_size)
            if not buffer:
                break
            sha256_hash.update(buffer)
    return sha256_hash.hexdigest()


def _fetch_remote(url, file_path, file_checksum=None):
    """
    Fetch the wanted dataset from the given url and save it in file_path.

    Parameters
    ----------
    url : string
        The url to fetch the dataset from.
    file_path : string
        Full path of the downloaded file including filename.
    file_checksum : string
        The file checksum using sha256 to check against the one computed on the downloaded file.
        Default is 'None', which means the checksum is not checked.

    Raises
    ------
    OSError
        If the computed SHA256 checksum of file does not match the one given by the user.
    """

    # Get the file
    urlretrieve(url, file_path)

    if file_checksum is not None:
        checksum = _checksum_sha256(file_path)
        if file_checksum != checksum:
            # Remove file and raise error
            remove(file_path)
            raise OSError(
                "{} has a SHA256 checksum : {}, "
                "different from expected : {}."
                "The file may be corrupted or the given url may be wrong !".format(file_path, checksum, file_checksum)
            )


def _fetch_remote_license(license_url, license_path, license_checksum=None, accept_license=False):
    """
    Fetch the wanted license from the given url and save it in file_path.

    Parameters
    ----------
    license_url : string
        The url to fetch the license file from.
    license_path : string
        Full path of the downloaded file including filename.
    license_checksum : string
        The license file checksum using sha256 to check against the one computed on the downloaded file.
        Default is 'None', which means the checksum is not checked.

    Raises
    ------
    OSError
        If the computed SHA256 checksum of file does not match the one given by the user.
    """
    _fetch_remote(license_url, license_path, license_checksum)
    # Print license terms unless accept_license is set to True
    if not accept_license:
        if exists(license_path):
            with open(license_path) as f:
                print(f.read())


def _get_archive_path(file_path, label):
    """
    Get archive path based on file_path given by user and label.

    Parameters
    ----------
    file_path: string
        Full path of the file to get including filename, or None.
    label: string
        Label used along with 'data_home' to get archive path, in case 'file_path' is None.

    Returns
    -------
        Full path of archive including filename.
    """
    if file_path is None:
        archive_path = join(_get_data_home(), label)
        dirname = split(archive_path)[0]
        makedirs(dirname, exist_ok=True)
    else:
        archive_path = file_path
        dirname = split(archive_path)[0]
        makedirs(dirname, exist_ok=True)

    return archive_path


def fetch_spiral_2d(file_path=None):
    """
    Load the spiral_2d dataset.

    Note that if the dataset already exists in the target location, it is not downloaded again,
    and the corresponding array is returned from cache.

    Parameters
    ----------
    file_path : string
        Full path of the downloaded file including filename.

        Default is None, meaning that it's set to "data_home/points/spiral_2d/spiral_2d.npy".

        The "data_home" directory is set by default to "~/gudhi_data",
        unless the 'GUDHI_DATA' environment variable is set.

    Returns
    -------
    points: numpy array
        Array of shape (114562, 2).
    """
    file_url = "https://raw.githubusercontent.com/GUDHI/gudhi-data/main/points/spiral_2d/spiral_2d.npy"
    file_checksum = "2226024da76c073dd2f24b884baefbfd14928b52296df41ad2d9b9dc170f2401"

    archive_path = _get_archive_path(file_path, "points/spiral_2d/spiral_2d.npy")

    if not exists(archive_path):
        _fetch_remote(file_url, archive_path, file_checksum)

    return np.load(archive_path, mmap_mode="r")


def fetch_bunny(file_path=None, accept_license=False):
    """
    Load the Stanford bunny dataset.

    This dataset contains 35947 vertices.

    Note that if the dataset already exists in the target location, it is not downloaded again,
    and the corresponding array is returned from cache.

    Parameters
    ----------
    file_path : string
        Full path of the downloaded file including filename.

        Default is None, meaning that it's set to "data_home/points/bunny/bunny.npy".
        In this case, the LICENSE file would be downloaded as "data_home/points/bunny/bunny.LICENSE".

        The "data_home" directory is set by default to "~/gudhi_data",
        unless the 'GUDHI_DATA' environment variable is set.

    accept_license : boolean
        Flag to specify if user accepts the file LICENSE and prevents from printing the corresponding license terms.

        Default is False.

    Returns
    -------
    points: numpy array
        Array of shape (35947, 3).
    """

    file_url = "https://raw.githubusercontent.com/GUDHI/gudhi-data/main/points/bunny/bunny.npy"
    file_checksum = "f382482fd89df8d6444152dc8fd454444fe597581b193fd139725a85af4a6c6e"
    license_url = "https://raw.githubusercontent.com/GUDHI/gudhi-data/main/points/bunny/bunny.LICENSE"
    license_checksum = "b763dbe1b2fc6015d05cbf7bcc686412a2eb100a1f2220296e3b4a644c69633a"

    archive_path = _get_archive_path(file_path, "points/bunny/bunny.npy")

    if not exists(archive_path):
        _fetch_remote(file_url, archive_path, file_checksum)
        license_path = join(split(archive_path)[0], "bunny.LICENSE")
        _fetch_remote_license(license_url, license_path, license_checksum, accept_license)

    return np.load(archive_path, mmap_mode="r")


@lru_cache(maxsize=None)
def _load_and_cache_activity(file_path):
    return np.load(file_path, mmap_mode="r")


def fetch_daily_activities(file_path=None, subset=None, accept_license=False):
    """
    Load a subset of the Daily and Sports Activities dataset. This dataset comes from
    https://archive.ics.uci.edu/ml/datasets/daily+and+sports+activities (CC BY 4.0 license).

    Note that if the dataset already exists in the target location, it is not downloaded again,
    and the corresponding dataset is read from cache.

    Parameters
    ----------
    file_path : string
        Full path of the downloaded file including filename.

        Default is None, meaning that it's set to "data_home/points/activities/activities_p1_left_leg.npy".
        In this case, the LICENSE file would be downloaded as "data_home/points/activities/activities.LICENSE".

        The "data_home" directory is set by default to "~/gudhi_data",
        unless the 'GUDHI_DATA' environment variable is set.

    subset : string
        This argument allows to download the following subsets:
         * 'cross_training' Only left leg magnetometer of cross training activity performed by the person 1.
           It contains 7.500 vertices in dimension 3.
         * 'jumping' Only left leg magnetometer of jumping activity performed by the person 1.
           It contains 7.500 vertices in dimension 3.
         * 'stepper' Only left leg magnetometer of stepper activity performed by the person 1.
           It contains 7.500 vertices in dimension 3.
         * 'walking' Only left leg magnetometer of walking activity performed by the person 1.
           It contains 7.500 vertices in dimension 3.
         * None (default value) This dataset contains 30.000 vertices in dimension 3 + activity type column
           (`14.` for 'cross_training', `18.` for 'jumping', `13.` for 'stepper', or `9.` for 'walking')

    accept_license : boolean
        Flag to specify if user accepts the file LICENSE and prevents from printing the corresponding license terms.

        Default is False.

    Returns
    -------
    points: numpy array
        Depending on subset value:
         * Array of shape (7.500, 3, dtype = float).

        Or
         * Array of shape (30.000, 4, dtype = float) when `subset is None`.
    """
    # To avoid download when subset is not correct
    if subset not in ["walking", "stepper", "cross_training", "jumping", None]:
        raise ValueError("Unknown subset value")

    file_url = "https://raw.githubusercontent.com/GUDHI/gudhi-data/main/points/activities/activities_p1_left_leg.npy"
    file_checksum = "ff813f717dbd3c8f8a95e59a7d8496d0b43c440e85a4db1f2b338cbfa9c02f25"
    activities_license_url = (
        "https://raw.githubusercontent.com/GUDHI/gudhi-data/main/points/activities/activities.LICENSE"
    )
    activities_license_checksum = "f5ce6749fa9d5359d7b0c4c37d0b61e5d9520f9494cd53be94295d3967ee4023"
    gudhi_data_set_path = "points/activities/activities_p1_left_leg.npy"

    archive_path = _get_archive_path(file_path, gudhi_data_set_path)
    if not exists(archive_path):
        _fetch_remote(file_url, archive_path, file_checksum)
        license_path = join(split(archive_path)[0], "activities.LICENSE")
        _fetch_remote_license(activities_license_url, license_path, activities_license_checksum, accept_license)

    ds = _load_and_cache_activity(archive_path)

    if subset is not None:
        activity_pos = {"walking": 0, "stepper": 1, "cross_training": 2, "jumping": 3}
        per_activity = 7500
        start_pos = activity_pos[subset] * per_activity
        ds = ds[start_pos : (start_pos + per_activity), 0:3]

    return ds



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Vincent Rouvreau
#
# Copyright (C) 2016 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from cython cimport numeric
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp.string cimport string
from libcpp cimport bool
from libc.stdint cimport intptr_t

from gudhi.simplex_tree cimport *
from gudhi.simplex_tree import SimplexTree

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

cdef extern from "Rips_complex_interface.h" namespace "Gudhi":
    cdef cppclass Rips_complex_interface "Gudhi::rips_complex::Rips_complex_interface":
        Rips_complex_interface() nogil
        void init_points(vector[vector[double]] values, double threshold) nogil
        void init_matrix(vector[vector[double]] values, double threshold) nogil
        void init_points_sparse(vector[vector[double]] values, double threshold, double sparse) nogil
        void init_matrix_sparse(vector[vector[double]] values, double threshold, double sparse) nogil
        void create_simplex_tree(Simplex_tree_python_interface* simplex_tree, int dim_max) nogil except +

# RipsComplex python interface
cdef class RipsComplex:
    """The data structure is a one skeleton graph, or Rips graph, containing edges when the edge length is less or
    equal to a given threshold. Edge length is computed from a user given point cloud with a given distance function,
    or a distance matrix.
    """

    cdef Rips_complex_interface thisref

    # Fake constructor that does nothing but documenting the constructor
    def __init__(self, *, points=None, distance_matrix=None, max_edge_length=float('inf'), sparse=None):
        """RipsComplex constructor.

        :param points: A list of points in d-Dimension.
        :type points: List[List[float]]

        Or

        :param distance_matrix: A distance matrix (full square or lower triangular).
        :type distance_matrix: List[List[float]]

        And in both cases

        :param max_edge_length: Maximal edge length. All edges of the graph strictly greater than `threshold` are not
            inserted in the graph.
        :type max_edge_length: float
        :param sparse: If this is not None, it switches to building a sparse Rips and represents the approximation
            parameter epsilon.
        :type sparse: float
        """

    # The real cython constructor
    def __cinit__(self, *, points=None, distance_matrix=None, max_edge_length=float('inf'), sparse=None):
        if sparse is not None:
          if distance_matrix is not None:
              self.thisref.init_matrix_sparse(distance_matrix, max_edge_length, sparse)
          else:
              if points is None:
                  # Empty Rips construction
                  points=[]
              self.thisref.init_points_sparse(points, max_edge_length, sparse)
        else:
          if distance_matrix is not None:
              self.thisref.init_matrix(distance_matrix, max_edge_length)
          else:
              if points is None:
                  # Empty Rips construction
                  points=[]
              self.thisref.init_points(points, max_edge_length)


    def create_simplex_tree(self, max_dimension=1):
        """
        :param max_dimension: graph expansion for Rips until this given maximal dimension.
        :type max_dimension: int
        :returns: A simplex tree encoding the Vietoris–Rips filtration.
        :rtype: SimplexTree
        """
        stree = SimplexTree()
        cdef intptr_t stree_int_ptr=stree.thisptr
        cdef int maxdim = max_dimension
        with nogil:
            self.thisref.create_simplex_tree(<Simplex_tree_python_interface*>stree_int_ptr, maxdim)
        return stree



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Vincent Rouvreau
#
# Copyright (C) 2022 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from .. import RipsComplex
from sklearn.base import BaseEstimator, TransformerMixin

# joblib is required by scikit-learn
from joblib import Parallel, delayed

# Mermaid sequence diagram - https://mermaid-js.github.io/mermaid-live-editor/
# sequenceDiagram
#   participant USER
#   participant R as RipsPersistence
#   USER->>R: fit_transform(X)
#   Note right of R: homology_dimensions=[i,j]
#   R->>thread1: _tranform(X[0])
#   R->>thread2: _tranform(X[1])
#   Note right of R: ...
#   thread1->>R: [array( Hi(X[0]) ), array( Hj(X[0]) )]
#   thread2->>R: [array( Hi(X[1]) ), array( Hj(X[1]) )]
#   Note right of R: ...
#   R->>USER: [[array( Hi(X[0]) ), array( Hj(X[0]) )],<br/> [array( Hi(X[1]) ), array( Hj(X[1]) )],<br/>...]


class RipsPersistence(BaseEstimator, TransformerMixin):
    """
    This is a class for constructing Vietoris-Rips complexes and computing the persistence diagrams from them.
    """

    def __init__(
        self,
        homology_dimensions,
        threshold=float('inf'),
        input_type='point cloud',
        num_collapses=1,
        homology_coeff_field=11,
        n_jobs=None,
    ):
        """
        Constructor for the RipsPersistence class.

        Parameters:
            homology_dimensions (int or list of int): The returned persistence diagrams dimension(s).
                Short circuit the use of :class:`~gudhi.representations.preprocessing.DimensionSelector` when only one
                dimension matters (in other words, when `homology_dimensions` is an int).
            threshold (float): Rips maximal edge length value. Default is +Inf.
            input_type (str): Can be 'point cloud' when inputs are point clouds, or 'lower distance matrix', when
                inputs are lower triangular distance matrix (can be full square, but the upper part of the distance
                matrix will not be considered). Default is 'point cloud'.
            num_collapses (int): Specify the number of :func:`~gudhi.SimplexTree.collapse_edges` iterations to perform
                on the SimplexTree. Default value is 1 (a relatively good enough number of iterations).
            homology_coeff_field (int): The homology coefficient field. Must be a prime number. Default value is 11.
            n_jobs (int): Number of jobs to run in parallel. `None` (default value) means `n_jobs = 1` unless in a
                joblib.parallel_backend context. `-1` means using all processors. cf.
                https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html for more details.
        """
        self.homology_dimensions = homology_dimensions
        self.threshold = threshold
        self.input_type = input_type
        self.num_collapses = num_collapses
        self.homology_coeff_field = homology_coeff_field
        self.n_jobs = n_jobs

    def fit(self, X, Y=None):
        """
        Nothing to be done, but useful when included in a scikit-learn Pipeline.
        """
        return self

    def __transform(self, inputs):
        max_dimension = max(self.dim_list_) + 1

        if self.input_type == 'point cloud':
            rips = RipsComplex(points=inputs, max_edge_length = self.threshold)
        elif self.input_type == 'lower distance matrix':
            rips = RipsComplex(distance_matrix=inputs, max_edge_length = self.threshold)
        else:
            raise ValueError("Only 'point cloud' and  'lower distance matrix' are valid input_type")
        
        if max_dimension > 1:
            stree = rips.create_simplex_tree(max_dimension=1)
            stree.collapse_edges(nb_iterations = self.num_collapses)
            stree.expansion(max_dimension)
        else:
            stree = rips.create_simplex_tree(max_dimension=max_dimension)

        persistence_dim_max = False
        # Specific case where, despite expansion(max_dimension), stree has a lower dimension
        if max_dimension > stree.dimension():
            persistence_dim_max = True

        stree.compute_persistence(
            homology_coeff_field=self.homology_coeff_field,
            persistence_dim_max=persistence_dim_max
        )

        return [
            stree.persistence_intervals_in_dimension(dim) for dim in self.dim_list_
        ]

    def transform(self, X, Y=None):
        """Compute all the Vietoris-Rips complexes and their associated persistence diagrams.

        :param X: list of point clouds as Euclidean coordinates or distance matrices.
        :type X: list of list of float OR list of numpy.ndarray

        :return: Persistence diagrams in the format:

              - If `homology_dimensions` was set to `n`: `[array( Hn(X[0]) ), array( Hn(X[1]) ), ...]` 
              - If `homology_dimensions` was set to `[i, j]`:
                `[[array( Hi(X[0]) ), array( Hj(X[0]) )], [array( Hi(X[1]) ), array( Hj(X[1]) )], ...]`
        :rtype: list of numpy ndarray of shape (,2) or list of list of numpy ndarray of shape (,2)
        """
        # Depends on homology_dimensions is an integer or a list of integer (else case)
        if isinstance(self.homology_dimensions, int):
            unwrap = True
            self.dim_list_ = [ self.homology_dimensions ]
        else:
            unwrap = False
            self.dim_list_ = self.homology_dimensions

        # threads is preferred as Rips construction and persistence computation releases the GIL
        res = Parallel(n_jobs=self.n_jobs, prefer="threads")(delayed(self.__transform)(inputs) for inputs in X)

        if unwrap:
            res = [d[0] for d in res]
        return res



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Vincent Rouvreau
#
# Copyright (C) 2016 Inria
#
# Modification(s):
#   - YYYY/MM Author: Description of the modification

from cython cimport numeric
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool
import os

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT (GPL v3 for sparsify_point_set)"

cdef extern from "Subsampling_interface.h" namespace "Gudhi::subsampling":
    vector[vector[double]] subsampling_n_farthest_points(bool, vector[vector[double]] points, size_t nb_points)
    vector[vector[double]] subsampling_n_farthest_points(bool, vector[vector[double]] points, size_t nb_points, size_t starting_point)
    vector[vector[double]] subsampling_n_farthest_points_from_file(bool, string off_file, size_t nb_points)
    vector[vector[double]] subsampling_n_farthest_points_from_file(bool, string off_file, size_t nb_points, size_t starting_point)
    vector[vector[double]] subsampling_n_random_points(vector[vector[double]] points, unsigned nb_points)
    vector[vector[double]] subsampling_n_random_points_from_file(string off_file, unsigned nb_points)
    vector[vector[double]] subsampling_sparsify_points(vector[vector[double]] points, double min_squared_dist)
    vector[vector[double]] subsampling_sparsify_points_from_file(string off_file, double min_squared_dist)

cdef extern from "Subsampling_interface.h":
    cdef int _GUDHI_SUBSAMPLING_USE_CGAL

GUDHI_SUBSAMPLING_USE_CGAL = _GUDHI_SUBSAMPLING_USE_CGAL

def choose_n_farthest_points(points=None, off_file='', nb_points=-<size_t>1, starting_point=None, fast=True):
    """Subsample by a greedy strategy of iteratively adding the farthest point
    from the current chosen point set to the subsampling.
    The iteration starts with the landmark `starting point`.

    :param points: The input point set.
    :type points: Iterable[Iterable[float]]

    Or

    :param off_file: An OFF file style name.
    :type off_file: string

    And in both cases

    :param nb_points: Number of points of the subsample (the subsample may be
        smaller if there are fewer than nb_points distinct input points). Default: all of them.
    :type nb_points: int
    :param starting_point: The iteration starts with the landmark `starting
        point`, which is the index of the point to start with. If not set, this
        index is chosen randomly.
    :type starting_point: int
    :param fast: If True (default), use an implementation that is efficient when the doubling dimension
        and spread are small, but slow otherwise. If False, use the standard quadratic algorithm.
    :type fast: bool
    :returns:  The subsample point set, in the order they were selected by the greedy strategy.
    :rtype: List[List[float]]
    """
    if off_file:
        if os.path.isfile(off_file):
            if starting_point is None:
                return subsampling_n_farthest_points_from_file(fast, off_file.encode('utf-8'),
                                                               nb_points)
            else:
                return subsampling_n_farthest_points_from_file(fast, off_file.encode('utf-8'),
                                                               nb_points, starting_point)
        else:
            print("file " + off_file + " not found.")
    else:
        if points is None:
            # Empty points
            points=[]
        if starting_point is None:
            return subsampling_n_farthest_points(fast, points, nb_points)
        else:
            return subsampling_n_farthest_points(fast, points, nb_points, starting_point)

def pick_n_random_points(points=None, off_file='', nb_points=0):
    """Subsample a point set by picking random vertices.

    :param points: The input point set.
    :type points: Iterable[Iterable[float]]

    Or

    :param off_file: An OFF file style name.
    :type off_file: string

    And in both cases

    :param nb_points: Number of points of the subsample.
    :type nb_points: int
    :returns:  The subsample point set.
    :rtype: List[List[float]]
    """
    if off_file:
        if os.path.isfile(off_file):
            return subsampling_n_random_points_from_file(off_file.encode('utf-8'),
                nb_points)
        else:
            print("file " + off_file + " not found.")
    else:
        if points is None:
            # Empty points
            points=[]
        return subsampling_n_random_points(points, nb_points)

def sparsify_point_set(points=None, off_file='', min_squared_dist=0.0):
    """Outputs a subset of the input points so that the squared distance
    between any two points is greater than min_squared_dist.

    :param points: The input point set.
    :type points: Iterable[Iterable[float]]

    Or

    :param off_file: An OFF file style name.
    :type off_file: string

    And in both cases

    :param min_squared_dist: Minimum squared distance separating the output \
    points.
    :type min_squared_dist: float
    :returns:  The subsample point set.
    :rtype: List[List[float]]
    """
    if not GUDHI_SUBSAMPLING_USE_CGAL:
        raise NotImplementedError("subsampling sparsify_point_set is only available with CGAL >= 4.11 and Eigen3")
    
    if off_file:
        if os.path.isfile(off_file):
            return subsampling_sparsify_points_from_file(off_file.encode('utf-8'),
                                                         min_squared_dist)
        else:
            print("file " + off_file + " not found.")
    else:
        if points is None:
            # Empty points
            points=[]
        return subsampling_sparsify_points(points, min_squared_dist)



# This file is part of the Gudhi Library - https://gudhi.inria.fr/ - which is released under MIT.
# See file LICENSE or go to https://gudhi.inria.fr/licensing/ for full license details.
# Author(s):       Vincent Rouvreau
#
# Copyright (C) 2016 Inria
#
# Modification(s):
#   - 2023/02 Vincent Rouvreau: Add serialize/deserialize for pickle feature
#   - YYYY/MM Author: Description of the modification

from cython cimport numeric
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from libcpp cimport bool
from libcpp.string cimport string

__author__ = "Vincent Rouvreau"
__copyright__ = "Copyright (C) 2016 Inria"
__license__ = "MIT"

cdef extern from "Simplex_tree_interface.h" namespace "Gudhi":
    cdef cppclass Simplex_tree_simplex_handle "Gudhi::Simplex_tree_interface::Simplex_handle":
        pass

    cdef cppclass Simplex_tree_simplices_iterator "Gudhi::Simplex_tree_interface::Complex_simplex_iterator":
        Simplex_tree_simplices_iterator() nogil
        Simplex_tree_simplex_handle& operator*() nogil
        Simplex_tree_simplices_iterator operator++() nogil
        bint operator!=(Simplex_tree_simplices_iterator) nogil

    cdef cppclass Simplex_tree_skeleton_iterator "Gudhi::Simplex_tree_interface::Skeleton_simplex_iterator":
        Simplex_tree_skeleton_iterator() nogil
        Simplex_tree_simplex_handle& operator*() nogil
        Simplex_tree_skeleton_iterator operator++() nogil
        bint operator!=(Simplex_tree_skeleton_iterator) nogil

    cdef cppclass Simplex_tree_boundary_iterator "Gudhi::Simplex_tree_interface::Boundary_simplex_iterator":
        Simplex_tree_boundary_iterator() nogil
        Simplex_tree_simplex_handle& operator*() nogil
        Simplex_tree_boundary_iterator operator++() nogil
        bint operator!=(Simplex_tree_boundary_iterator) nogil


    cdef cppclass Simplex_tree_python_interface "Gudhi::Simplex_tree_interface":
        Simplex_tree_python_interface() nogil
        Simplex_tree_python_interface(Simplex_tree_python_interface&) nogil
        double simplex_filtration(vector[int] simplex) nogil
        void assign_simplex_filtration(vector[int] simplex, double filtration) nogil except +
        void initialize_filtration() nogil
        int num_vertices() nogil
        int num_simplices() nogil
        bool is_empty() nogil
        void set_dimension(int dimension) nogil
        int dimension() nogil
        int upper_bound_dimension() nogil
        bool find_simplex(vector[int] simplex) nogil
        bool insert(vector[int] simplex, double filtration) nogil
        void insert_matrix(double* filtrations, int n, int stride0, int stride1, double max_filtration) nogil except +
        void insert_batch_vertices(vector[int] v, double f) nogil except +
        vector[pair[vector[int], double]] get_star(vector[int] simplex) nogil
        vector[pair[vector[int], double]] get_cofaces(vector[int] simplex, int dimension) nogil
        void expansion(int max_dim) nogil except +
        void remove_maximal_simplex(vector[int] simplex) nogil
        bool prune_above_filtration(double filtration) nogil
        bool prune_above_dimension(int dimension) nogil
        bool make_filtration_non_decreasing() nogil
        void compute_extended_filtration() nogil
        void collapse_edges(int nb_collapse_iteration) nogil except +
        void reset_filtration(double filtration, int dimension) nogil
        bint operator==(Simplex_tree_python_interface) nogil
        # Iterators over Simplex tree
        pair[vector[int], double] get_simplex_and_filtration(Simplex_tree_simplex_handle f_simplex) nogil
        Simplex_tree_simplices_iterator get_simplices_iterator_begin() nogil
        Simplex_tree_simplices_iterator get_simplices_iterator_end() nogil
        vector[Simplex_tree_simplex_handle].const_iterator get_filtration_iterator_begin() nogil
        vector[Simplex_tree_simplex_handle].const_iterator get_filtration_iterator_end() nogil
        Simplex_tree_skeleton_iterator get_skeleton_iterator_begin(int dimension) nogil
        Simplex_tree_skeleton_iterator get_skeleton_iterator_end(int dimension) nogil
        pair[Simplex_tree_boundary_iterator, Simplex_tree_boundary_iterator] get_boundary_iterators(vector[int] simplex) nogil except +
        # Expansion with blockers
        ctypedef bool (*blocker_func_t)(vector[int], void *user_data) except +
        void expansion_with_blockers_callback(int dimension, blocker_func_t user_func, void *user_data) except +
        void serialize(char* buffer, const size_t buffer_size) nogil except +
        void deserialize(const char* buffer, const size_t buffer_size) nogil except +
        size_t get_serialization_size() nogil
        void clear() nogil

cdef extern from "Persistent_cohomology_interface.h" namespace "Gudhi":
    cdef cppclass Simplex_tree_persistence_interface "Gudhi::Persistent_cohomology_interface<Gudhi::Simplex_tree_interface>":
        Simplex_tree_persistence_interface(Simplex_tree_python_interface * st, bool persistence_dim_max) nogil
        void compute_persistence(int homology_coeff_field, double min_persistence) nogil except +
        vector[pair[int, pair[double, double]]] get_persistence() nogil
        vector[int] betti_numbers() nogil
        vector[int] persistent_betti_numbers(double from_value, double to_value) nogil
        vector[pair[double,double]] intervals_in_dimension(int dimension) nogil
        void write_output_diagram(string diagram_file_name) nogil except +
        vector[pair[vector[int], vector[int]]] persistence_pairs() nogil
        pair[vector[vector[int]], vector[vector[int]]] lower_star_generators() nogil
        pair[vector[vector[int]], vector[vector[int]]] flag_generators() nogil
        vector[vector[pair[int, pair[double, double]]]] compute_extended_persistence_subdiagrams(double min_persistence) nogil
